import asyncio
import json
import time
import traceback
import uuid
from typing import AsyncGenerator, Optional

import httpx

import src.chat as ch
import src.format as format
import src.log as log
from src.config import BOOST_PUBLIC_URL, EXTRA_LLM_PARAMS, INTERMEDIATE_OUTPUT
from src.events import AsyncEventEmitter
from src.llm_registry import llm_registry

# import src.mods as mods
# import src.tools as tools
# import src.tools.registry as tools_registry

logger = log.setup_logger(__name__)

BOOST_PARAM_PREFIX = "@boost_"


class LLM(AsyncEventEmitter):
    url: str
    headers: dict
    query_params: dict

    model: str
    params: dict
    boost_params: dict
    module: str
    provider: str
    raw_payload: Optional[dict]

    queue: asyncio.Queue
    is_streaming: bool
    is_final_stream: bool

    cpl_id: int

    def __init__(self, **kwargs):
        super().__init__()

        self.id = str(uuid.uuid4())
        self.url = kwargs.get("url")
        self.headers = kwargs.get("headers", {})
        self.query_params = kwargs.get("query_params", {})

        self.model = kwargs.get("model")
        self.split_params(kwargs.get("params", {}))

        self.chat = self.resolve_chat(**kwargs)
        self.messages = self.chat.history()

        self.module = kwargs.get("module")
        self.provider = kwargs.get("provider", "openai")
        self.raw_payload = kwargs.get("raw_payload")

        self.queue = asyncio.Queue()
        self.queues = []
        self.is_streaming = False
        self.is_final_stream = False

        self.cpl_id = 0

    @property
    def chat_completion_endpoint(self):
        return f"{self.url}/chat/completions"

    def split_params(self, params: dict):
        self.params = {
            k: v
            for k, v in {**EXTRA_LLM_PARAMS.value, **params}.items()
            if not k.startswith(BOOST_PARAM_PREFIX)
        }
        self.boost_params = {
            k[len(BOOST_PARAM_PREFIX) :]: v
            for k, v in params.items()
            if k.startswith(BOOST_PARAM_PREFIX)
        }

    def generate_system_fingerprint(self):
        return "fp_boost"

    def generate_chunk_id(self):
        self.cpl_id += 1
        return f"chatcmpl-{self.cpl_id}"

    def get_response_content(self, params: dict, response: dict):
        content = response["choices"][0]["message"]["content"]

        if "response_format" in params and "type" in params["response_format"]:
            if (
                params["response_format"]["type"] == "json_schema"
                or params["response_format"]["type"] == "json"
            ):
                return json.loads(content)

        return content

    def inspect_completion(
        self,
        *,
        system_prompt: str,
        user_prompt: str,
        model: Optional[str] = None,
        response_format: Optional[dict] = None,
    ):
        provider = getattr(self, "provider", "openai")
        if provider == "anthropic":
            return self._inspect_completion_anthropic(
                system_prompt=system_prompt,
                user_prompt=user_prompt,
                model=model,
                response_format=response_format,
            )

        return self._inspect_completion_openai(
            system_prompt=system_prompt,
            user_prompt=user_prompt,
            model=model,
            response_format=response_format,
        )

    def _inspect_completion_openai(
        self,
        *,
        system_prompt: str,
        user_prompt: str,
        model: Optional[str] = None,
        response_format: Optional[dict] = None,
    ):
        resolved_model = model or self.model
        if not resolved_model:
            raise RuntimeError("missing model for inspection completion")

        params = dict(self.params or {})
        params.pop("stream", None)
        params.pop("stream_options", None)
        params.pop("messages", None)

        body = {
            "model": resolved_model,
            "messages": [
                {"role": "system", "content": system_prompt},
                {"role": "user", "content": user_prompt},
            ],
            **params,
            "stream": False,
        }

        if response_format is not None:
            body["response_format"] = response_format

        with httpx.Client(timeout=30.0) as client:
            response = client.post(
                self.chat_completion_endpoint,
                headers=self.headers,
                params=self.query_params,
                json=body,
            )
            response.raise_for_status()
            payload = response.json()

        return self._extract_inspection_text(payload)

    def _inspect_completion_anthropic(
        self,
        *,
        system_prompt: str,
        user_prompt: str,
        model: Optional[str] = None,
        response_format: Optional[dict] = None,
    ):
        resolved_model = model or self.model
        if not resolved_model:
            raise RuntimeError("missing model for inspection completion")

        body = {
            "model": resolved_model,
            "system": system_prompt,
            "messages": [{"role": "user", "content": user_prompt}],
            "max_tokens": 256,
        }

        if response_format is not None:
            body["response_format"] = response_format

        endpoint = self._provider_endpoint("/messages")

        with httpx.Client(timeout=30.0) as client:
            response = client.post(
                endpoint,
                headers=self.headers,
                params=self.query_params,
                json=body,
            )
            response.raise_for_status()
            payload = response.json()

        return self._extract_inspection_text(payload)

    def _provider_endpoint(self, path: str):
        base = (self.url or "").rstrip("/")
        normalized = path if path.startswith("/") else f"/{path}"

        if getattr(self, "provider", "openai") == "anthropic":
            if normalized.startswith("/v1/"):
                return f"{base}{normalized}"
            if base.endswith("/v1"):
                return f"{base}{normalized}"
            return f"{base}/v1{normalized}"

        return f"{base}{normalized}"

    def _extract_inspection_text(self, payload: dict):
        if not isinstance(payload, dict):
            raise ValueError("invalid inspector response payload")

        choices = payload.get("choices")
        if isinstance(choices, list) and choices and isinstance(choices[0], dict):
            message = choices[0].get("message", {})
            content = message.get("content")
            return self._content_to_text(content)

        content_blocks = payload.get("content")
        if isinstance(content_blocks, list):
            text_parts = []
            for block in content_blocks:
                if isinstance(block, dict) and isinstance(block.get("text"), str):
                    text_parts.append(block["text"])
            if text_parts:
                return "\n".join(text_parts).strip()

        raise ValueError("invalid inspector response: missing text content")

    def _content_to_text(self, content):
        if isinstance(content, str):
            return content.strip()

        if isinstance(content, list):
            text_parts = []
            for part in content:
                if isinstance(part, dict) and isinstance(part.get("text"), str):
                    text_parts.append(part["text"])
            if text_parts:
                return "\n".join(text_parts).strip()

        raise ValueError("invalid inspector response: missing text content")

    def get_chunk_content(self, chunk):
        try:
            choices = chunk.get("choices", [])
            choice = choices[0] if choices and len(choices) > 0 else {}
            delta = choice.get("delta", {})
            return delta.get("content", "")
        except (KeyError, IndexError):
            logger.error(f"Unexpected chunk format: {chunk}")
            return ""

    def get_chunk_tool_calls(self, chunk):
        try:
            choices = chunk.get("choices", [])
            choice = choices[0] if choices and len(choices) > 0 else {}
            delta = choice.get("delta", {})
            return delta.get("tool_calls", [])
        except (KeyError, IndexError):
            logger.error(f"Unexpected chunk format: {chunk}")
            return

    def parse_chunk(self, chunk):
        if isinstance(chunk, dict):
            return chunk

        if isinstance(chunk, bytes):
            chunk = chunk.decode("utf-8")

        chunk_str = chunk.split("\n")[0]
        if chunk_str.startswith("data: "):
            chunk_str = chunk_str[6:]

        if chunk_str == "[DONE]":
            return self.chunk_from_message("")

        return json.loads(chunk_str)

    def output_from_chunk(self, chunk):
        return {
            "id": chunk["id"],
            "object": "chat.completion",
            "created": chunk["created"],
            "model": self.model,
            "system_fingerprint": self.generate_system_fingerprint(),
            "choices": [
                {
                    "index": choice["index"],
                    "message": {
                        "role": choice["delta"].get("role", "assistant"),
                        "content": choice["delta"].get("content", ""),
                    },
                    "finish_reason": None,
                }
                for choice in chunk["choices"]
            ],
            "usage": {"prompt_tokens": 0, "completion_tokens": 0, "total_tokens": 0},
        }

    def chunk_from_delta(self, delta: dict):
        now = int(time.time())

        return {
            "id": self.generate_chunk_id(),
            "object": "chat.completion.chunk",
            "created": now,
            "model": self.model,
            "system_fingerprint": self.generate_system_fingerprint(),
            "choices": [{"index": 0, "delta": delta, "finish_reason": None}],
        }

    def chunk_from_message(self, message: str):
        return self.chunk_from_delta({"role": "assistant", "content": message})

    def chunk_from_tool_call(self, tool_call: dict):
        if "index" not in tool_call:
            tool_call["index"] = 0

        return self.chunk_from_delta({"role": "assistant", "tool_calls": [tool_call]})

    def chunk_to_string(self, chunk):
        if isinstance(chunk, dict):
            chunk = f"data: {json.dumps(chunk)}\n\n"

        return chunk

    def is_tool_call(self, chunk):
        choices = chunk.get("choices", [])
        choice = choices[0] if choices and len(choices) > 0 else {}
        delta = choice.get("delta", {})
        has_tool_calls = delta.get("tool_calls", [])
        return len(has_tool_calls) > 0

    def event_to_string(self, event, data):
        payload = {"object": "boost.listener.event", "event": event, "data": data}

        return f"data: {json.dumps(payload)}\n\n"

    async def serve(self):
        logger.debug("Serving boosted LLM...")
        llm_registry.register(self)

        async def apply_mod():
            try:
                if self.module is None:
                    logger.debug("No module specified")
                    await self.stream_final_completion()
                    return

                # mod = mods.registry.get(self.module)
                # if mod is None:
                #   logger.error(f"Module '{self.module}' not found.")
                #   return
                logger.warning("Modules are currently disabled.")
                await self.stream_final_completion()
            except httpx.HTTPStatusError as e:
                logger.error(f"Upstream HTTP error: {e}")
                try:
                    content = e.response.content.decode("utf-8")
                    try:
                        error_json = json.loads(content)
                        error_data = error_json.get("error", {"message": content})
                    except Exception:
                        error_data = {"message": content}
                except Exception:
                    error_data = {"message": str(e)}

                await self.emit_data(
                    json.dumps(
                        {
                            "error": {
                                "message": error_data.get("message", str(e)),
                                "type": error_data.get("type", "upstream_error"),
                                "code": e.response.status_code,
                            }
                        }
                    )
                )
            except httpx.ConnectError as e:
                await self.emit_data(
                    json.dumps(
                        {
                            "error": {
                                "message": f"Failed to connect to downstream API: {str(e)}",
                                "type": "connection_error",
                                "code": 502,
                            }
                        }
                    )
                )
            except Exception as e:
                logger.error(f"Error in LLM service: {e}")
                # Emit generic error if needed
            finally:
                await self.emit_done()

            # logger.debug(f"Applying '{self.module}' to '{self.model}'")
            # try:
            #   self.chat.llm = self
            #   await mod.apply(chat=self.chat, llm=self)
            # except Exception as e:
            #   logger.error(f"Failed to apply module '{self.module}': {e}")
            #   for line in traceback.format_tb(e.__traceback__):
            #     logger.error(line)

            # logger.debug(f"'{self.module}' application complete for '{self.model}'")
            # await self.emit_done()

        asyncio.create_task(apply_mod())
        return self.response_stream()

    async def generator(self):
        self.is_streaming = True

        while self.is_streaming or not self.queue.empty():
            chunk = await self.queue.get()

            if chunk is None:
                break

            yield chunk

    async def response_stream(self):
        async for chunk in self.generator():
            # Final stream is always passed back as
            # that's the useful payload of a given iteration
            if INTERMEDIATE_OUTPUT.value or self.is_final_stream:
                yield chunk

    async def listen(self):
        queue = asyncio.Queue()
        self.queues.append(queue)

        while True:
            chunk = await queue.get()
            if chunk is None:
                break
            yield chunk

    async def emit_status(self, status):
        await self.emit_message(format.format_status(status))

    async def emit_artifact(self, artifact, wait=True):
        artifact = artifact.replace("<<boost_public_url>>", BOOST_PUBLIC_URL.value).replace(
            "<<listener_id>>", self.id
        )
        await self.emit_message(format.format_artifact(artifact))
        if wait:
            await asyncio.sleep(1.0)

    async def emit_message(self, message):
        await self.emit_chunk(self.chunk_from_message(message))

    async def emit_chunk(self, chunk):
        if (
            "choices" not in chunk
            or not chunk["choices"]
            or "delta" not in chunk["choices"][0]
            or "content" not in chunk["choices"][0]["delta"]
        ):
            if "choices" not in chunk or not chunk["choices"]:
                chunk["choices"] = [{}]
            if "delta" not in chunk["choices"][0]:
                chunk["choices"][0]["delta"] = {}
            chunk["choices"][0]["delta"]["content"] = ""

        if "choices" not in chunk or not chunk["choices"] or "index" not in chunk["choices"][0]:
            if "choices" not in chunk or not chunk["choices"]:
                chunk["choices"] = [{}]
            chunk["choices"][0]["index"] = 0

        await self.emit_data(self.chunk_to_string(chunk))

    async def emit_data(self, data):
        await self.queue.put(data)
        await self.emit_to_listeners(data)

    async def emit_to_listeners(self, data):
        for queue in self.queues:
            await queue.put(data)

    async def emit_listener_event(self, event, data):
        await self.emit_to_listeners(self.event_to_string(event, data))

    async def emit_done(self):
        await self.emit_data("data: [DONE]")
        await self.emit_data(None)
        await self.remove_all_listeners()
        self.is_streaming = False

    async def stream_final_completion(self, **kwargs):
        self.is_final_stream = True
        return await self.stream_chat_completion(**kwargs)

    async def stream_chat_completion(self, **kwargs):
        request = await self.resolve_request(**kwargs)

        chat = request.get("chat", self.chat)
        params = request.get("params", self.params)
        model = request.get("model", self.model)
        url = request.get("url", self.url)
        headers = request.get("headers", self.headers)
        query_params = request.get("query_params", self.query_params)
        should_emit = kwargs.get("emit", True)

        logger.debug(f"Params: {params}")
        logger.debug(f"Chat: {str(chat):.256}")

        result = ""
        pending_tool_calls = {}  # Track tool calls being built, keyed by id
        first_tool_call_id = None

        async with httpx.AsyncClient(timeout=None) as client:
            current_stream_content = ""

            while True:
                # Assistant must remember what it said so far
                if current_stream_content and current_stream_content != "":
                    if chat.tail.role == "assistant":
                        chat.tail.content += current_stream_content
                    else:
                        chat.assistant(current_stream_content)

                body = {
                    "model": model,
                    "messages": chat.history(),
                    **params,
                    "stream": True,
                    "stream_options": {
                        "include_usage": True,
                    },
                }

                logger.info(body)

                # Flag to determine if we need to execute tool calls
                end_of_stream = False
                current_stream_content = ""
                # Track tool call IDs in order of appearance
                tool_call_order = []

                async with client.stream(
                    "POST",
                    f"{url}/chat/completions",
                    headers=headers,
                    params=query_params,
                    json=body,
                ) as response:
                    if response.status_code >= 400:
                        body = await response.aread()
                        logger.error(
                            f"Chat completion error {response.status_code}: {body.decode('utf-8')}"
                        )
                        # Ensure we raise an error that will be caught by the serving loop
                        response.raise_for_status()

                    buffer = b""

                    async for chunk in response.aiter_bytes():
                        buffer += chunk

                        while b"\n" in buffer:
                            line, buffer = buffer.split(b"\n", 1)
                            line = line.decode("utf-8").strip()

                            if not line or line.startswith(":"):
                                continue

                            if line == "data: [DONE]":
                                end_of_stream = True
                                continue

                            if not line.startswith("data:"):
                                continue

                            try:
                                parsed = self.parse_chunk(line)

                                # Safely check finish_reason
                                choices = parsed.get("choices", [])
                                if choices and len(choices) > 0:
                                    finish_reason = choices[0].get("finish_reason")
                                    if finish_reason == "tool_calls":
                                        end_of_stream = True

                                # Extract content for regular text responses
                                content = self.get_chunk_content(parsed)
                                if content:
                                    current_stream_content += content
                                    result += content

                                # Process tool call chunks
                                if self.is_tool_call(parsed):
                                    # Extract tool call data safely
                                    choices = parsed.get("choices", [])
                                    if not choices:
                                        continue

                                    delta = choices[0].get("delta", {})
                                    tool_calls_data = delta.get("tool_calls", [])

                                    if not tool_calls_data:
                                        continue

                                    tool_call = tool_calls_data[0]
                                    tool_id = tool_call.get("id")
                                    index = tool_call.get("index", 0)

                                    # Store the first tool call ID we see
                                    if tool_id and not first_tool_call_id:
                                        first_tool_call_id = tool_id

                                    # Use tool_id as primary key if available, fall back to index
                                    # This handles Ollama bug where multiple calls
                                    # have same index but different ids
                                    if tool_id:
                                        key = tool_id
                                    else:
                                        # For streaming chunks without id,
                                        # find existing call by index
                                        # or use index as key for truly streamed arguments
                                        key = f"idx_{index}"
                                        for (
                                            existing_key,
                                            existing_call,
                                        ) in pending_tool_calls.items():
                                            if existing_call.get("_index") == index:
                                                key = existing_key
                                                break

                                    # Initialize tool call if new
                                    if key not in pending_tool_calls:
                                        pending_tool_calls[key] = {
                                            "id": tool_id or first_tool_call_id,
                                            "function": {
                                                "name": tool_call.get("function", {}).get("name"),
                                                "arguments": "",
                                            },
                                            "type": tool_call.get("type") or "function",
                                            "_index": index,
                                        }
                                        tool_call_order.append(key)

                                    # Update arguments
                                    function_args = tool_call.get("function", {}).get("arguments")
                                    if key in pending_tool_calls and function_args is not None:
                                        pending_tool_calls[key]["function"]["arguments"] += (
                                            function_args
                                        )

                                    logger.debug(f"Tool call chunk: {parsed}")
                                else:
                                    if should_emit:
                                        await self.emit_chunk(parsed)

                            except json.JSONDecodeError:
                                logger.error(f'Failed to parse chunk: "{line}"')
                            except Exception as e:
                                logger.error(f"Error processing chunk: {str(e)}")
                                for line in traceback.format_tb(e.__traceback__):
                                    logger.error(line)

                # After stream ends, check if we need to execute tool calls
                if pending_tool_calls and (end_of_stream or not current_stream_content):
                    for key in tool_call_order:
                        tool_call = pending_tool_calls.get(key)
                        if not tool_call:
                            continue
                        # Remove internal tracking field before use
                        tool_call.pop("_index", None)

                        logger.info(f"Passing back to API client: {tool_call}")
                        await self.emit_chunk(self.chunk_from_tool_call(tool_call))

                    return result

                # If no tool calls or content was streamed, we're done
                break

        return result

    async def chat_completion(self, **kwargs):
        chat = self.resolve_chat(**kwargs)
        params = await self.resolve_request_params(**kwargs)
        should_resolve = kwargs.get("resolve", False)

        logger.debug(f"Chat Completion for '{self.chat_completion_endpoint}'")
        logger.debug(f"Params: {params}")
        logger.debug(f"Chat: {str(chat):.256}...")
        if chat is None:
            chat = self.chat

        async with httpx.AsyncClient(timeout=None) as client:
            body = {"model": self.model, "messages": chat.history(), **params, "stream": False}
            response = await client.post(
                self.chat_completion_endpoint, headers=self.headers, json=body
            )
            result = response.json()
            if should_resolve:
                return self.get_response_content(params, result)
            return result

    async def consume_stream(self, stream: AsyncGenerator[bytes, None]):
        output_obj = None
        content = ""
        tool_calls = []

        async for chunk_bytes in stream:
            chunk = self.parse_chunk(chunk_bytes)
            if output_obj is None:
                output_obj = self.output_from_chunk(chunk)
            chunk_content = self.get_chunk_content(chunk)
            chunk_tools = self.get_chunk_tool_calls(chunk)

            content += chunk_content
            tool_calls.extend(chunk_tools)

        if output_obj:
            output_obj["choices"][0]["message"]["content"] = content

            if len(tool_calls) > 0:
                output_obj["choices"][0]["message"]["tool_calls"] = tool_calls
                output_obj["choices"][0]["finish_reason"] = "tool_calls"

        return output_obj

    async def resolve_request_params(self, **kwargs):
        params = {
            "model": kwargs.get("model", self.model),
            **self.params,
            **kwargs.get("params", {}),
        }

        # local_tools = tools.registry.collect_tool_defs()
        local_tools = []

        if local_tools:
            params["tools"] = params.get("tools", [])
            params["tools"].extend(local_tools)

        if kwargs.get("schema"):
            params["response_format"] = {
                "type": "json_schema",
                "json_schema": {
                    "name": "StructuredResponseSchema",
                    "schema": kwargs["schema"].model_json_schema(),
                },
            }

        return params

    def resolve_chat(
        self,
        messages: Optional[list] = None,
        chat: Optional["ch.Chat"] = None,
        prompt: Optional[str] = None,
        **prompt_kwargs,
    ):
        if chat is not None:
            return chat

        if messages is not None:
            return ch.Chat.from_conversation(messages)

        if prompt is not None:
            message = prompt.format(**prompt_kwargs)
            return ch.Chat.from_conversation([{"role": "user", "content": message}])

        if hasattr(self, "chat"):
            return self.chat

        return ch.Chat()

    async def resolve_model(self, model: Optional[str] = None, **rest) -> str:
        return model or self.model

    async def resolve_headers(self, **kwargs):
        return self.headers

    async def resolve_query_params(self, **kwargs):
        return self.query_params

    async def resolve_url(self, **kwargs):
        return self.url

    async def resolve_request(self, **kwargs):
        logger.debug("resolving")

        tasks = {
            "url": self.resolve_url(**kwargs),
            "headers": self.resolve_headers(**kwargs),
            "params": self.resolve_request_params(**kwargs),
            "model": self.resolve_model(**kwargs),
            "query_params": self.resolve_query_params(**kwargs),
        }

        values = await asyncio.gather(*tasks.values())
        results = {k: v for k, v in zip(tasks.keys(), values)}
        results["chat"] = self.resolve_chat(**kwargs)

        return results

    async def start_thinking(self):
        await self.emit_message("\n<think>\n")

    async def stop_thinking(self):
        await self.emit_message("\n</think>\n")



import random
import re
from typing import TYPE_CHECKING, Any, Dict, List, TypeVar, Union

from src import dotty, log

if TYPE_CHECKING:
    from chat import Chat

logger = log.setup_logger(__name__)
T = TypeVar("T")


class BooleanExpressionNode:
    pass


class TerminalNode(BooleanExpressionNode):
    def __init__(self, term: str):
        self.term = term.strip()


class AndNode(BooleanExpressionNode):
    def __init__(self, left: BooleanExpressionNode, right: BooleanExpressionNode):
        self.left = left
        self.right = right


class OrNode(BooleanExpressionNode):
    def __init__(self, left: BooleanExpressionNode, right: BooleanExpressionNode):
        self.left = left
        self.right = right


def select_percentage(chat: "Chat", **kwargs):
    percentage = kwargs.get("percentage", 50)
    nodes = chat.plain()
    num_nodes = max(1, int(len(nodes) * (percentage / 100)))

    return nodes[:num_nodes]


def select_match(chat: "Chat", **kwargs):
    substring = kwargs.get("substring", "")
    role = kwargs.get("role", "")
    index = kwargs.get("index", None)

    nodes = chat.plain()

    if role:
        nodes = [node for node in nodes if node.role == role]

    if substring:
        nodes = [node for node in nodes if substring in node.content]

    if index is not None:
        nodes = [nodes[index]]

    return nodes


def select_from_user(chat: "Chat"):
    return select_match(chat, role="user")


def select_all(chat: "Chat"):
    return chat.plain()


def select_first(chat: "Chat"):
    return select_match(chat, index=0)


def select_last(chat: "Chat"):
    return select_match(chat, index=-1)


def select_any(chat: "Chat"):
    return [random.choice(chat.plain())]


selection_strategies = {
    "all": select_all,
    "first": select_first,
    "last": select_last,
    "any": select_any,
    "percentage": select_percentage,
    "match": select_match,
    "user": select_from_user,
}


def apply_strategy(chat: "Chat", strategy: str, params: dict):
    return selection_strategies[strategy](chat, **params)  # type: ignore


def match_regex(value, regex):
    return bool(re.match(regex, value))


def match_substring(value, substring):
    return substring in value


def match_exact(value, target):
    return value == target


def filter_array_elements(items: List[T], filter_obj: Dict[str, Any]) -> List[T]:
    """
    Filter elements of an array based on a filter object.

    Args:
        array: The array to filter
        filter_obj: A Hasura-style filter object

    Returns:
        list: Filtered array
    """
    return [item for item in items if match_filter(item, filter_obj)]


def find_array_element(items: List[T], filter_obj: Dict[str, Any]) -> Union[T, None]:
    """
    Find the first element in an array that matches a filter object.

    Args:
        array: The array to search
        filter_obj: A Hasura-style filter object
    Returns:
        Any: The first matching element, or None if no match is found
    """

    for item in items:
        if match_filter(item, filter_obj):
            return item
    return None


def match_filter(context: Any, filter_obj: Dict[str, Any]):
    """
    Match a context object (dict, class instance, etc.) against a Hasura-style filter.

    Args:
        context: The object to match against the filter
        filter_obj: A Hasura-style filter object

    Returns:
        bool: True if the context matches the filter, False otherwise
    """
    if filter_obj is None:
        return True

    # Handle empty filter
    if not filter_obj:
        return True

    # Convert class instance to dict if needed
    if not isinstance(context, dict):
        context = {
            k: getattr(context, k)
            for k in dir(context)
            if not k.startswith("_") and not callable(getattr(context, k))
        }

    # Process logical operators
    if "_and" in filter_obj:
        return all(match_filter(context, sub_filter) for sub_filter in filter_obj["_and"])

    if "_or" in filter_obj:
        return any(match_filter(context, sub_filter) for sub_filter in filter_obj["_or"])

    if "_not" in filter_obj:
        return not match_filter(context, filter_obj["_not"])

    # Process field conditions
    for field, conditions in filter_obj.items():
        if field in ("_and", "_or", "_not"):
            continue

        # Get field value, treating missing fields as None
        if dotty.has(context, field):
            current = dotty.get(context, field)
        else:
            current = None

        # Handle both direct operators and nested conditions
        if isinstance(conditions, dict):
            if any(k.startswith("_") for k in conditions.keys()):
                # Direct operators
                for operator, expected_value in conditions.items():
                    # For _is_null, we allow missing fields to be treated as None
                    if operator == "_is_null":
                        if not apply_operator(current, operator, expected_value):
                            return False
                    else:
                        # For other operators, missing fields should fail the match
                        if current is None and not dotty.has(context, field):
                            return False
                        if not apply_operator(current, operator, expected_value):
                            return False
            else:
                # Nested conditions
                if not match_filter(current, conditions):
                    return False
        else:
            # Direct value comparison
            if not match_exact(current, conditions):
                return False

    return True


def apply_operator(field_value, operator, expected_value):
    """Apply a Hasura operator to compare field_value with expected_value."""

    # Equal operators
    if operator == "_eq":
        return field_value == expected_value
    if operator == "_neq":
        return field_value != expected_value
    if operator == "_is_null":
        return (field_value is None) == expected_value

    # Comparison operators
    if operator == "_gt":
        return field_value > expected_value
    if operator == "_lt":
        return field_value < expected_value
    if operator == "_gte":
        return field_value >= expected_value
    if operator == "_lte":
        return field_value <= expected_value
    if operator == "_neq":
        return field_value != expected_value

    # Array operators
    if operator == "_in":
        return field_value in expected_value
    if operator == "_nin":
        return field_value not in expected_value
    if operator == "_some":
        if isinstance(field_value, list):
            return any(match_filter(item, expected_value) for item in field_value)
        return False
    if operator == "_every":
        if isinstance(field_value, list):
            return all(match_filter(item, expected_value) for item in field_value)
        return False
    if operator == "_none":
        if isinstance(field_value, list):
            return not any(match_filter(item, expected_value) for item in field_value)
        return True
    if operator == "_is_empty":
        if isinstance(field_value, list):
            return (len(field_value) == 0) == expected_value
        elif isinstance(field_value, dict):
            return (len(field_value) == 0) == expected_value
        return field_value == []

    # String operators
    if operator in ("_like", "_ilike", "_nlike", "_nilike"):
        pattern_value = str(expected_value).replace("%", "\x00")
        escaped_value = re.escape(pattern_value)
        pattern = escaped_value.replace("\x00", ".*")
        is_case_insensitive = "ilike" in operator
        is_negated = operator.startswith("_n")

        try:
            flags = re.DOTALL
            if is_case_insensitive:
                flags |= re.IGNORECASE

            result = bool(re.match(f"^{pattern}$", str(field_value), flags))

            return not result if is_negated else result
        except re.error:
            return False

    # Regex operators
    if operator in ("_regex", "_iregex", "_nregex", "_niregex"):
        is_case_insensitive = "iregex" in operator
        is_negated = operator.startswith("_n")

        try:
            if is_case_insensitive:
                result = bool(re.search(str(expected_value), str(field_value), re.IGNORECASE))
            else:
                result = bool(re.search(str(expected_value), str(field_value)))

            return not result if is_negated else result
        except re.error:
            return False

    # JSON operators
    if operator == "_contains":
        if isinstance(field_value, dict) and isinstance(expected_value, dict):
            # Check if all key-value pairs in expected_value exist in field_value
            return all(k in field_value and field_value[k] == v for k, v in expected_value.items())
        elif isinstance(field_value, list):
            # Check if expected_value is in the list
            return expected_value in field_value
        elif isinstance(expected_value, list) and not isinstance(field_value, (dict, list)):
            # If expected_value is a list but field_value is a scalar
            return field_value in expected_value
        else:
            # Direct comparison for other types
            return field_value == expected_value

    raise ValueError(f"Unsupported operator: {operator}")


def _tokenize_boolean_expression(text: str) -> List[str]:
    import re

    pattern = r"(\(|\)|AND|OR)"
    parts = re.split(pattern, text)

    tokens = []
    for part in parts:
        stripped = part.strip()
        if stripped:
            if stripped in ("(", ")", "AND", "OR"):
                tokens.append(stripped)
            else:
                normalized = " ".join(stripped.split())
                if normalized:
                    tokens.append(normalized)

    open_count = tokens.count("(")
    close_count = tokens.count(")")
    if open_count > close_count:
        raise ValueError("Unbalanced parentheses: missing closing parenthesis")
    elif close_count > open_count:
        raise ValueError("Unbalanced parentheses: too many closing parentheses")

    return tokens


def _parse_or_expression(tokens: List[str], pos: int) -> tuple[BooleanExpressionNode, int]:
    left, pos = _parse_and_expression(tokens, pos)

    while pos < len(tokens) and tokens[pos] == "OR":
        pos += 1
        right, pos = _parse_and_expression(tokens, pos)
        left = OrNode(left, right)

    return left, pos


def _parse_and_expression(tokens: List[str], pos: int) -> tuple[BooleanExpressionNode, int]:
    left, pos = _parse_primary_expression(tokens, pos)

    while pos < len(tokens) and tokens[pos] == "AND":
        pos += 1
        right, pos = _parse_primary_expression(tokens, pos)
        left = AndNode(left, right)

    return left, pos


def _parse_primary_expression(tokens: List[str], pos: int) -> tuple[BooleanExpressionNode, int]:
    if pos >= len(tokens):
        raise ValueError("Unexpected end of expression")

    token = tokens[pos]

    if token == "(":
        pos += 1
        expr, pos = _parse_or_expression(tokens, pos)
        if pos >= len(tokens) or tokens[pos] != ")":
            raise ValueError("Expected closing parenthesis")
        return expr, pos + 1

    if token in ("AND", "OR", ")"):
        raise ValueError(f"Unexpected token: {token}")

    return TerminalNode(token), pos + 1


def parse_boolean_expression(text: str) -> Union[str, BooleanExpressionNode]:
    if not text or not text.strip():
        raise ValueError("Expression cannot be empty")

    text = text.strip()

    if "AND" not in text and "OR" not in text:
        return text

    tokens = _tokenize_boolean_expression(text)

    if not tokens:
        raise ValueError("Expression cannot be empty")

    expr, pos = _parse_or_expression(tokens, 0)

    if pos < len(tokens):
        raise ValueError(f"Unexpected token after expression: {tokens[pos]}")

    return expr


def build_search_filter(expression: str, field_builder_fn) -> Dict[str, Any]:
    parsed = parse_boolean_expression(expression)

    if isinstance(parsed, str):
        return field_builder_fn(parsed)

    def convert_ast(node: BooleanExpressionNode) -> Dict[str, Any]:
        if isinstance(node, TerminalNode):
            return field_builder_fn(node.term)
        elif isinstance(node, AndNode):
            return {"_and": [convert_ast(node.left), convert_ast(node.right)]}
        elif isinstance(node, OrNode):
            return {"_or": [convert_ast(node.left), convert_ast(node.right)]}
        else:
            raise ValueError(f"Unknown node type: {type(node)}")

    return convert_ast(parsed)



"""Content filtering guard - blocks specific words/phrases."""

import re
from typing import TYPE_CHECKING, Dict, List

if TYPE_CHECKING:
    from src.chat import Chat
    from src.llm import LLM


def apply(chat: "Chat", llm: "LLM", config: Dict) -> List[str]:
    """
    Filter blocked words from message content.

    Args:
        chat: Chat object to filter
        llm: LLM object
        config: Guard configuration with 'blocked_words' list

    Returns:
        List of audit logs
    """
    blocked_words = config.get("blocked_words", [])
    if not blocked_words:
        return []

    audit_logs = []

    # Iterate over all messages in the chat
    # chat.plain() returns list of ChatNodes
    for idx, node in enumerate(chat.plain()):
        content = node.content

        if content is None:
            continue

        if isinstance(content, str):
            modified = False
            for word in blocked_words:
                # Case-insensitive search
                if word.lower() in content.lower():
                    pattern = re.compile(re.escape(word), re.IGNORECASE)
                    content = pattern.sub("[FILTERED]", content)
                    modified = True
                    audit_logs.append(
                        f"content_filter: Replaced '{word}' in message {idx} ({node.role})"
                    )

            if modified:
                node.content = content

        elif isinstance(content, list):
            # Multimodal content
            for part_idx, part in enumerate(content):
                if isinstance(part, dict) and "text" in part:
                    original_text = part["text"]
                    text = original_text
                    modified = False

                    for word in blocked_words:
                        if word.lower() in text.lower():
                            pattern = re.compile(re.escape(word), re.IGNORECASE)
                            text = pattern.sub("[FILTERED]", text)
                            modified = True
                            audit_logs.append(
                                f"content_filter: Replaced '{word}' "
                                f"in message {idx}, part {part_idx} ({node.role})"
                            )

                    if modified:
                        part["text"] = text

    return audit_logs



from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from src import llm


class LLMRegistry:
    def __init__(self):
        self._registry = {}

    def register(self, instance: "llm.LLM"):
        self._registry[instance.id] = instance

    def get(self, llm_id) -> "llm.LLM":
        return self._registry.get(llm_id)

    def unregister(self, instance: "llm.LLM"):
        del self._registry[instance.id]

    def list_all(self):
        return list(self._registry.values())


llm_registry = LLMRegistry()



"""
OpenGuard - Main FastAPI Application

This module creates the FastAPI application that handles:
- Health checks
- Model listing
- Chat completions with guard application
- Authentication and request middleware
"""

import json

import httpx
from fastapi import Depends, FastAPI, HTTPException, Request
from fastapi.middleware.cors import CORSMiddleware
from fastapi.responses import JSONResponse, Response, StreamingResponse

from src import config, llm, mapper
from src import log as log_module
from src.chat import Chat
from src.guard_engine import apply_guards, log_audit
from src.guards import GuardBlockedError, get_guards
from src.middleware.request_id import RequestIDMiddleware
from src.middleware.request_state import RequestStateMiddleware

# Setup logging
logger = log_module.setup_logger(__name__)

# Create FastAPI app
app = FastAPI(title="OpenGuard", description="guarding proxy for AI", version="0.1.0")

# Add middlewares in correct order
app.add_middleware(RequestStateMiddleware)
app.add_middleware(RequestIDMiddleware)
app.add_middleware(
    CORSMiddleware,
    allow_origins=config.OPENGUARD_CORS_ORIGINS.value,
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)


async def verify_auth(request: Request):
    """Verify API key if configured"""
    api_key = config.OPENGUARD_API_KEY.value
    api_keys_list = config.OPENGUARD_API_KEYS.value

    # If no keys configured, allow all requests
    if not api_key and not api_keys_list:
        return True

    # Check Authorization header
    auth_header = request.headers.get("Authorization", "")
    if not auth_header.startswith("Bearer "):
        raise HTTPException(status_code=401, detail="Missing or invalid Authorization header")

    token = auth_header[7:]  # Remove "Bearer "

    # Check against configured keys
    valid_keys = [api_key] if api_key else []
    valid_keys.extend(api_keys_list)

    if token not in valid_keys:
        raise HTTPException(status_code=403, detail="Invalid API key")

    return True


def _anthropic_error(status_code: int, message: str, error_type: str = "invalid_request_error"):
    return JSONResponse(
        status_code=status_code,
        content={"type": "error", "error": {"type": error_type, "message": message}},
    )


def _is_guard_blocked_error(error: Exception):
    return error.__class__.__name__ == "GuardBlockedError"


def _requires_explicit_anthropic_tool_use(payload: dict | None):
    if not isinstance(payload, dict):
        return False

    if payload.get("stream"):
        return False

    tools = payload.get("tools")
    tool_choice = payload.get("tool_choice")

    if not isinstance(tools, list) or len(tools) == 0:
        return False

    return isinstance(tool_choice, dict) and tool_choice.get("type") == "tool"


def _anthropic_message_has_tool_use(response_payload: dict | None):
    if not isinstance(response_payload, dict):
        return False

    content_blocks = response_payload.get("content")
    if not isinstance(content_blocks, list):
        return False

    return any(
        isinstance(block, dict) and block.get("type") == "tool_use" for block in content_blocks
    )


def _build_forward_headers(request: Request, provider: str, route_config: dict):
    headers = {
        key: value
        for key, value in request.headers.items()
        if key.lower() not in {"host", "content-length"}
    }

    mode = route_config.get("mode")
    downstream_key = route_config.get("key")

    if mode in {"internal", "endpoint"} and downstream_key:
        if provider == "anthropic":
            headers.pop("authorization", None)
            headers.pop("Authorization", None)
            headers["x-api-key"] = downstream_key
        else:
            headers.pop("x-api-key", None)
            headers.pop("X-API-Key", None)
            headers["Authorization"] = f"Bearer {downstream_key}"

    return headers


class _GuardContext:
    def __init__(self, model, params, provider, raw_payload):
        self.model = model
        self.params = params
        self.provider = provider
        self.raw_payload = raw_payload


async def _forward_provider_request(
    request: Request,
    provider: str,
    endpoint_path: str,
    body_bytes: bytes | None,
    payload: dict | None,
):
    incoming_key = mapper.extract_api_key(request.headers)
    model = (payload or {}).get("model") if isinstance(payload, dict) else None

    if incoming_key and incoming_key in mapper.get_internal_api_keys() and model:
        model_map = mapper.MODEL_TO_BACKEND.get(provider.lower(), {})
        if model not in model_map:
            try:
                await mapper.list_downstream(provider=provider)
            except Exception as e:
                logger.warning(f"Unable to warm model cache for provider '{provider}': {e}")

    try:
        route_config = mapper.resolve_provider_route(
            provider=provider,
            endpoint_path=endpoint_path,
            body=payload,
            headers=request.headers,
        )
    except ValueError as e:
        return _anthropic_error(500, str(e), error_type="api_error")

    proxy_headers = _build_forward_headers(request, provider, route_config)
    query_params = dict(request.query_params)

    try:
        async with httpx.AsyncClient(timeout=None) as client:
            response = await client.request(
                method=request.method,
                url=route_config["endpoint"],
                headers=proxy_headers,
                params=query_params,
                content=body_bytes if body_bytes else None,
            )

            if (
                provider == "anthropic"
                and endpoint_path == "/messages/count_tokens"
                and response.status_code == 404
                and isinstance(payload, dict)
            ):
                fallback_payload = dict(payload)
                fallback_payload.setdefault("max_tokens", 1)
                fallback_payload["stream"] = False
                fallback_body = json.dumps(fallback_payload).encode("utf-8")

                fallback_route = mapper.resolve_provider_route(
                    provider="anthropic",
                    endpoint_path="/messages",
                    body=fallback_payload,
                    headers=request.headers,
                )
                fallback_headers = _build_forward_headers(request, "anthropic", fallback_route)

                fallback_response = await client.request(
                    method="POST",
                    url=fallback_route["endpoint"],
                    headers=fallback_headers,
                    params=query_params,
                    content=fallback_body,
                )

                if fallback_response.status_code < 400:
                    try:
                        fallback_json = fallback_response.json()
                    except json.JSONDecodeError:
                        fallback_json = {}

                    usage = fallback_json.get("usage") if isinstance(fallback_json, dict) else None
                    if isinstance(usage, dict) and usage.get("input_tokens") is not None:
                        return JSONResponse(
                            status_code=200,
                            content={"input_tokens": usage["input_tokens"]},
                        )

            if (
                provider == "anthropic"
                and endpoint_path == "/messages"
                and response.status_code < 400
                and _requires_explicit_anthropic_tool_use(payload)
            ):
                try:
                    downstream_payload = response.json()
                except json.JSONDecodeError:
                    downstream_payload = None

                if not _anthropic_message_has_tool_use(downstream_payload):
                    return _anthropic_error(
                        422,
                        (
                            "Downstream model did not return tool_use for explicit tool_choice; "
                            "tool calling may be unsupported by the selected model/backend."
                        ),
                        error_type="invalid_request_error",
                    )

        content_type = response.headers.get("content-type", "application/json")
        response_headers = {
            key: value
            for key, value in response.headers.items()
            if key.lower()
            not in {
                "content-length",
                "transfer-encoding",
                "connection",
                "content-type",
            }
        }

        return Response(
            content=response.content,
            status_code=response.status_code,
            media_type=content_type,
            headers=response_headers,
        )
    except httpx.ConnectError as e:
        logger.error(f"Anthropic proxy connection error: {e}")
        return _anthropic_error(502, f"Failed to connect to downstream API: {str(e)}", "api_error")
    except httpx.TimeoutException as e:
        logger.error(f"Anthropic proxy timeout error: {e}")
        return _anthropic_error(504, "Request to downstream API timed out", "api_error")
    except Exception as e:
        logger.error(f"Anthropic proxy error: {e}", exc_info=True)
        return _anthropic_error(500, "Internal server error", "api_error")


async def _proxy_provider_passthrough(request: Request, provider: str, endpoint_path: str):
    body_bytes = await request.body()
    payload = None

    if body_bytes:
        try:
            payload = json.loads(body_bytes)
        except json.JSONDecodeError:
            return _anthropic_error(400, "Invalid JSON in request body")

    return await _forward_provider_request(request, provider, endpoint_path, body_bytes, payload)


async def _proxy_anthropic_messages_with_guards(request: Request):
    body_bytes = await request.body()
    payload = None

    if body_bytes:
        try:
            payload = json.loads(body_bytes)
        except json.JSONDecodeError:
            return _anthropic_error(400, "Invalid JSON in request body")

    if not isinstance(payload, dict):
        return _anthropic_error(400, "Invalid request format")

    try:
        chat = Chat.from_payload(payload, provider="anthropic")
    except Exception as e:
        logger.error(f"Invalid Anthropic payload format: {e}")
        return _anthropic_error(400, "Invalid request format")

    param_keys_excluded = {"model", "messages", "system"}
    guard_params = {key: value for key, value in payload.items() if key not in param_keys_excluded}

    try:
        route_config = mapper.resolve_provider_route(
            provider="anthropic",
            endpoint_path="/messages",
            body=payload,
            headers=request.headers,
        )
    except ValueError as e:
        return _anthropic_error(500, str(e), error_type="api_error")

    inspection_headers = _build_forward_headers(request, "anthropic", route_config)

    llm_instance = llm.LLM(
        url=route_config["url"],
        headers=inspection_headers,
        query_params=dict(request.query_params),
        model=payload.get("model"),
        params=guard_params,
        chat=chat,
    )
    llm_instance.provider = "anthropic"
    llm_instance.raw_payload = payload

    try:
        guards = get_guards()
        _, audit_logs = apply_guards(chat, llm_instance, guards)
        log_audit(audit_logs)
    except GuardBlockedError as e:
        return _anthropic_error(403, str(e), error_type="invalid_request_error")
    except Exception as e:
        if _is_guard_blocked_error(e):
            return _anthropic_error(403, str(e), error_type="invalid_request_error")
        raise

    forwarded_payload = chat.serialize("anthropic")

    for key, value in llm_instance.params.items():
        if key not in param_keys_excluded:
            forwarded_payload[key] = value

    for key in list(forwarded_payload.keys()):
        if key not in param_keys_excluded and key in payload and key not in llm_instance.params:
            del forwarded_payload[key]

    forwarded_body = json.dumps(forwarded_payload).encode("utf-8")
    return await _forward_provider_request(
        request,
        provider="anthropic",
        endpoint_path="/messages",
        body_bytes=forwarded_body,
        payload=forwarded_payload,
    )


def _select_models_provider(request: Request):
    incoming_key = mapper.extract_api_key(request.headers)
    if not incoming_key:
        return "openai"

    anthropic_backends = mapper.get_provider_backends("anthropic")
    if any(backend.get("key") and backend["key"] == incoming_key for backend in anthropic_backends):
        return "anthropic"

    return "openai"


@app.get("/")
async def root():
    """Root endpoint - API information"""
    return {
        "name": "OpenGuard",
        "version": "0.1.0",
        "description": "guarding proxy for AI",
        "endpoints": {"health": "/health", "models": "/v1/models", "chat": "/v1/chat/completions"},
    }


@app.get("/health")
async def health():
    """Health check endpoint"""
    return {"status": "ok"}


@app.get("/v1/models")
async def list_models(request: Request, authorized: bool = Depends(verify_auth)):
    """List all available models from downstream APIs"""
    try:
        provider = _select_models_provider(request)
        models = await mapper.list_downstream(provider=provider)
        return {"object": "list", "data": models}
    except Exception as e:
        logger.error(f"Error listing models: {e}")
        raise HTTPException(status_code=500, detail="Failed to list models")


@app.get("/v1/anthropic/models")
async def list_anthropic_models(request: Request):
    """Dedicated Anthropic-compatible models listing endpoint."""
    return await _proxy_provider_passthrough(request, provider="anthropic", endpoint_path="/models")


@app.post("/v1/messages")
async def anthropic_messages(request: Request):
    """Dedicated Anthropic messages endpoint with guard-aware passthrough semantics."""
    return await _proxy_anthropic_messages_with_guards(request)


@app.post("/v1/messages/count_tokens")
async def anthropic_count_tokens(request: Request):
    """Dedicated Anthropic count_tokens endpoint if supported by downstream."""
    return await _proxy_provider_passthrough(
        request,
        provider="anthropic",
        endpoint_path="/messages/count_tokens",
    )


@app.post("/v1/chat/completions")
async def chat_completions(request: Request, authorized: bool = Depends(verify_auth)):
    """Handle chat completion with guards applied"""
    try:
        # Parse request body
        payload = await request.json()

        # Validate messages
        messages = payload.get("messages")
        if not isinstance(messages, list) or len(messages) == 0:
            raise HTTPException(status_code=400, detail="'messages' must be a non-empty list")
        for m in messages:
            if not isinstance(m, dict) or "content" not in m or "role" not in m:
                raise HTTPException(status_code=400, detail="Invalid message format")
            if m.get("content") is None:
                raise HTTPException(status_code=400, detail="Message content cannot be null")

        # Validate types
        if "max_tokens" in payload and not isinstance(payload["max_tokens"], int):
            raise HTTPException(status_code=400, detail="'max_tokens' must be integer")
        if "temperature" in payload and not isinstance(payload["temperature"], (int, float)):
            raise HTTPException(status_code=400, detail="'temperature' must be number")

        # Resolve backend
        try:
            proxy_config = mapper.resolve_request_config(payload)
        except ValueError as e:
            logger.error(f"Failed to resolve backend: {e}")
            raise HTTPException(status_code=400, detail=str(e))

        # Create LLM proxy instance
        llm_instance = llm.LLM(
            url=proxy_config["url"],
            headers=proxy_config["headers"],
            model=payload.get("model"),
            params=proxy_config["params"],
            messages=payload.get("messages"),
        )

        # Apply guards
        guards = get_guards()
        # Guards modify chat/llm in-place
        _, audit_logs = apply_guards(llm_instance.chat, llm_instance, guards)
        log_audit(audit_logs)

        # Proxy request
        stream = payload.get("stream", False)

        if stream:
            return StreamingResponse(await llm_instance.serve(), media_type="text/event-stream")
        else:
            # Non-streaming: collect response and return as JSON
            full_content = ""
            tool_calls = []
            # Dictionary to aggregate tool calls by index: {index: tool_call_dict}
            pending_tool_calls = {}
            final_data = None
            finish_reason = None

            async for chunk in await llm_instance.serve():
                # Handle error chunks (raw JSON)
                if not chunk.startswith("data: "):
                    try:
                        err = json.loads(chunk)
                        if "error" in err:
                            # Map error code if present, default to 500
                            code = int(err["error"].get("code", 500))
                            if code == 404:
                                code = 400
                            return JSONResponse(content=err, status_code=code)
                    except (json.JSONDecodeError, ValueError):
                        pass
                    continue

                # Handle SSE chunks
                line = chunk.strip()
                if line == "data: [DONE]":
                    continue

                try:
                    if len(line) > 6:
                        data = json.loads(line[6:])  # Skip "data: "

                        # Store metadata from last valid chunk
                        final_data = data

                        # Accumulate content
                        choices = data.get("choices", [])
                        if choices:
                            delta = choices[0].get("delta", {})
                            content = delta.get("content", "")
                            if content:
                                full_content += content

                            # Accumulate tool calls with merging
                            if "tool_calls" in delta:
                                for tc in delta["tool_calls"]:
                                    idx = tc.get("index")
                                    # Fallback: if no index (e.g. from some backends),
                                    # append to list or treat as new
                                    # But standard OpenAI stream has index.
                                    # If llm.py sends full objects without index,
                                    # we can just append, but let's try to handle both.
                                    if idx is None:
                                        tool_calls.append(tc)
                                    else:
                                        if idx not in pending_tool_calls:
                                            pending_tool_calls[idx] = tc
                                        else:
                                            current = pending_tool_calls[idx]
                                            # Merge arguments
                                            if "function" in tc:
                                                if (
                                                    "name" in tc["function"]
                                                    and tc["function"]["name"]
                                                ):
                                                    if "function" not in current:
                                                        current["function"] = {}
                                                    current["function"]["name"] = tc["function"][
                                                        "name"
                                                    ]

                                                if (
                                                    "arguments" in tc["function"]
                                                    and tc["function"]["arguments"]
                                                ):
                                                    if "function" not in current:
                                                        current["function"] = {"arguments": ""}
                                                    if "arguments" not in current["function"]:
                                                        current["function"]["arguments"] = ""
                                                    current["function"]["arguments"] += tc[
                                                        "function"
                                                    ]["arguments"]

                                            # Merge other fields if present
                                            if "id" in tc and tc["id"]:
                                                current["id"] = tc["id"]
                                            if "type" in tc and tc["type"]:
                                                current["type"] = tc["type"]

                            if "finish_reason" in choices[0]:
                                finish_reason = choices[0]["finish_reason"]

                except json.JSONDecodeError:
                    continue

            if final_data:
                # Add aggregated tool calls to the list
                if pending_tool_calls:
                    # Sort by index to maintain order
                    sorted_idxs = sorted(pending_tool_calls.keys())
                    for idx in sorted_idxs:
                        tool_calls.append(pending_tool_calls[idx])

                # Construct non-streaming response format
                response_obj = final_data.copy()
                response_obj["object"] = "chat.completion"

                # Reconstruct message
                message: dict[str, object] = {
                    "role": "assistant",
                    "content": full_content if full_content else None,
                }

                if tool_calls:
                    message["tool_calls"] = tool_calls

                if "choices" in response_obj and response_obj["choices"]:
                    response_obj["choices"][0]["message"] = message

                    # Set finish reason if captured
                    if finish_reason:
                        response_obj["choices"][0]["finish_reason"] = finish_reason
                    elif tool_calls:
                        # Default to tool_calls if not explicitly set but tools are present
                        response_obj["choices"][0]["finish_reason"] = "tool_calls"

                    # Remove delta if present
                    if "delta" in response_obj["choices"][0]:
                        del response_obj["choices"][0]["delta"]

                return JSONResponse(content=response_obj, status_code=200)
            else:
                raise HTTPException(status_code=500, detail="Empty response from downstream")

    except GuardBlockedError as e:
        return JSONResponse(
            status_code=403,
            content={"error": {"message": str(e), "type": "guard_block", "code": 403}},
        )
    except httpx.HTTPStatusError as e:
        error_body = await e.response.aread()
        logger.error(f"Downstream error {e.response.status_code}: {error_body.decode('utf-8')}")

        if e.response.status_code == 404:
            return JSONResponse(
                status_code=400,
                content={
                    "error": {
                        "message": "Model not found or invalid",
                        "type": "invalid_request_error",
                        "code": 400,
                    }
                },
            )

        try:
            content = json.loads(error_body)
        except json.JSONDecodeError:
            content = {
                "error": {
                    "message": error_body.decode("utf-8"),
                    "type": "downstream_error",
                    "code": e.response.status_code,
                }
            }

        return JSONResponse(status_code=e.response.status_code, content=content)
    except httpx.ConnectError as e:
        logger.error(f"Connection error: {e}")
        return JSONResponse(
            status_code=502,
            content={
                "error": {
                    "message": f"Failed to connect to downstream API: {str(e)}",
                    "type": "connection_error",
                    "code": 502,
                }
            },
        )
    except httpx.TimeoutException as e:
        logger.error(f"Timeout error: {e}")
        return JSONResponse(
            status_code=504,
            content={
                "error": {
                    "message": "Request to downstream API timed out",
                    "type": "timeout_error",
                    "code": 504,
                }
            },
        )
    except HTTPException:
        raise
    except json.JSONDecodeError as e:
        logger.error(f"Invalid JSON in request: {e}")
        raise HTTPException(status_code=400, detail="Invalid JSON in request body")
    except (TypeError, KeyError, AttributeError) as e:
        logger.error(f"Invalid request format: {e}")
        raise HTTPException(status_code=400, detail="Invalid request format")
    except Exception as e:
        if _is_guard_blocked_error(e):
            return JSONResponse(
                status_code=403,
                content={"error": {"message": str(e), "type": "guard_block", "code": 403}},
            )
        logger.error(f"Error in chat completion: {e}", exc_info=True)
        raise HTTPException(status_code=500, detail="Internal server error")


def main():
    """
    Entry point for the application script.
    Delegates to the CLI application.
    """
    from src.cli import app

    app()



"""PII filtering guard - detects and replaces personally identifiable information."""

import re
from typing import TYPE_CHECKING, List

if TYPE_CHECKING:
    from src.chat import Chat
    from src.llm import LLM

# PII regex patterns
EMAIL_PATTERN = re.compile(r"\b[A-Za-z0-9._%+-]+@[A-Za-z0-9.-]+\.[A-Z|a-z]{2,}\b")
PHONE_PATTERN = re.compile(r"\b\d{3}[-.]?\d{3}[-.]?\d{4}\b")
SSN_PATTERN = re.compile(r"\b\d{3}-\d{2}-\d{4}\b")
CREDITCARD_PATTERN = re.compile(r"\b\d{4}[\s-]?\d{4}[\s-]?\d{4}[\s-]?\d{4}\b")

PII_PATTERNS = {
    "email": (EMAIL_PATTERN, "<protected:email>"),
    "phone": (PHONE_PATTERN, "<protected:phone>"),
    "ssn": (SSN_PATTERN, "<protected:ssn>"),
    "creditcard": (CREDITCARD_PATTERN, "<protected:creditcard>"),
}


def apply(chat: "Chat", llm: "LLM", config: dict) -> List[str]:
    """
    Filter PII from message content.

    Args:
        chat: Chat object
        llm: LLM object
        config: Guard configuration

    Returns:
        List of audit logs
    """
    audit_logs = []
    pii_found = False

    for idx, node in enumerate(chat.plain()):
        content = node.content

        if content is None:
            continue

        # Handle string content
        if isinstance(content, str):
            modified_content = content

            for pii_type, (pattern, replacement) in PII_PATTERNS.items():
                matches = pattern.findall(modified_content)
                if matches:
                    for match in matches:
                        audit_logs.append(
                            f"pii_filter: Found {pii_type} in message {idx} ({node.role}): {match}"
                        )
                        pii_found = True
                    modified_content = pattern.sub(replacement, modified_content)

            if modified_content != content:
                node.content = modified_content

        # Handle array content (multimodal messages)
        elif isinstance(content, list):
            # content for list is mutable?
            # Usually chat node content stores the list object.

            for part_idx, part in enumerate(content):
                if isinstance(part, dict) and "text" in part:
                    modified_text = part["text"]
                    part_modified = False

                    for pii_type, (pattern, replacement) in PII_PATTERNS.items():
                        matches = pattern.findall(modified_text)
                        if matches:
                            for match in matches:
                                audit_logs.append(
                                    f"pii_filter: Found {pii_type} "
                                    f"in message {idx} ({node.role}), part {part_idx}: {match}"
                                )
                                pii_found = True
                            modified_text = pattern.sub(replacement, modified_text)
                            part_modified = True

                    if part_modified:
                        part["text"] = modified_text

    # If any PII was found, inject a system message at the start
    if pii_found:
        chat.system(
            "Note: PII has been filtered from the following conversation. "
            "Protected items appear as <protected:type>."
        )

    return audit_logs



"""Guard configuration loading and validation."""

import os
from typing import Any, Dict, List

import yaml
from pydantic import BaseModel, Field

from src import log
from src.config import OPENGUARD_CONFIG

logger = log.setup_logger(__name__)


class GuardBlockedError(Exception):
    """Exception raised when a guard blocks a request."""

    pass


class GuardAction(BaseModel):
    """A single guard action to apply."""

    type: str
    config: Dict[str, Any] = Field(default_factory=dict)


class GuardRule(BaseModel):
    """A guard rule with match criteria and actions."""

    match: Dict[str, Any]
    apply: List[GuardAction]


class GuardsConfig(BaseModel):
    """Top-level guards configuration."""

    guards: List[GuardRule] = Field(default_factory=list)


# Module-level cache for loaded guards
_guards_cache: List[GuardRule] | None = None


def load_guards_config(config_path: str) -> GuardsConfig:
    """
    Load guards configuration from YAML file.

    Args:
        config_path: Path to YAML configuration file

    Returns:
        GuardsConfig object with parsed guards

    Raises:
        FileNotFoundError: If config file doesn't exist
        yaml.YAMLError: If YAML is invalid
        ValidationError: If config doesn't match schema
    """
    if not os.path.exists(config_path):
        raise FileNotFoundError(f"Guards configuration file not found: {config_path}")

    try:
        with open(config_path, "r") as f:
            raw_config = yaml.safe_load(f)

        if raw_config is None:
            raw_config = {}

        config = GuardsConfig(**raw_config)
        logger.info(f"Loaded {len(config.guards)} guard(s) from {config_path}")
        return config

    except yaml.YAMLError as e:
        logger.error(f"Invalid YAML in {config_path}: {e}")
        raise
    except Exception as e:
        logger.error(f"Failed to parse guards config: {e}")
        raise


def get_guards() -> List[GuardRule]:
    """
    Get all loaded guard rules (cached).

    Returns:
        List of GuardRule objects
    """
    global _guards_cache

    if _guards_cache is None:
        config_path = OPENGUARD_CONFIG.value
        try:
            config = load_guards_config(config_path)
            _guards_cache = config.guards
        except FileNotFoundError:
            logger.warning(f"No guards config found at {config_path}, using empty guards list")
            _guards_cache = []
        except Exception as e:
            logger.error(f"Error loading guards: {e}")
            _guards_cache = []

    return _guards_cache



import random
from typing import List, Optional

from . import log

logger = log.setup_logger(__name__)


class ChatNode:
    id: str
    content: str
    role: str
    tool_call_id: Optional[str]
    tool_calls: Optional[List[dict]]
    passthrough: dict

    parent: Optional["ChatNode"]
    children: List["ChatNode"]

    visits: int
    value: float
    meta: dict

    @staticmethod
    def from_conversation(messages):
        if not messages:
            return None

        root_message = messages[0]
        node = ChatNode(
            role=root_message.get("role", ""),
            content=root_message.get("content", ""),
            tool_call_id=root_message.get("tool_call_id"),
            tool_calls=root_message.get("tool_calls", []),
            passthrough={
                k: v
                for k, v in root_message.items()
                if k not in ["role", "content", "tool_call_id", "tool_calls"]
            },
        )

        for message in messages[1:]:
            child = ChatNode(
                role=message.get("role", ""),
                content=message.get("content", ""),
                tool_call_id=message.get("tool_call_id"),
                tool_calls=message.get("tool_calls", []),
                passthrough={
                    k: v
                    for k, v in message.items()
                    if k not in ["role", "content", "tool_call_id", "tool_calls"]
                },
            )
            node.add_child(child)
            node = child

        return node

    def __init__(self, **kwargs):
        self.id = "".join(random.choices("abcdefghijklmnopqrstuvwxyz0987654321", k=4))
        self.content = kwargs.get("content", "")
        self.role = kwargs.get("role", "")

        self.parent = kwargs.get("parent", None)
        self.children = kwargs.get("children", [])

        self.visits = kwargs.get("visits", 0)
        self.value = kwargs.get("value", 0.0)

        self.meta = kwargs.get("meta", {})

        self.tool_call_id = kwargs.get("tool_call_id", None)
        self.tool_calls = kwargs.get("tool_calls", [])
        self.passthrough = kwargs.get("passthrough", {})

    def add_parent(self, new_parent: "ChatNode"):
        # Guard against None and self-reference
        if new_parent is None or self == new_parent:
            return self

        # Remove from current parent if exists
        if self.parent:
            self.parent.children.remove(self)

        # Set new parent
        self.parent = new_parent
        if self not in new_parent.children:
            new_parent.children.append(self)

        return self

    # add child - similar to adding another branch
    # to the conversation from this node
    def add_child(self, child: "ChatNode"):
        # Guard against None and duplicates
        if child is None or child in self.children:
            return self

        for c in self.children:
            c.add_parent(child)

        # Update child's parent
        if child.parent:
            child.parent.children.remove(child)
        child.parent = self

        # Add to children
        self.children.append(child)
        return self

    # insert child - similar to adding a new node
    # in the middle of the conversation
    def insert_child(self, child: "ChatNode"):
        self.add_child(child)
        for c in self.children:
            c.add_parent(child)
        return self

    def remove_child(self, child: "ChatNode"):
        if child in self.children:
            self.children.remove(child)
            child.parent = None
        return self

    def best_child(self):
        if not self.children:
            return self
        return max(self.children, key=lambda c: c.value).best_child()

    def contains(self, substring):
        return substring.lower() in self.content.lower()

    def parents(self):
        parents = [self]

        while self.parent:
            self = self.parent
            parents.append(self)

        return parents[::-1]

    def message(self):
        result = {
            **self.passthrough,
            "role": self.role,
            "content": self.content,
        }

        if self.tool_call_id is not None:
            result["tool_call_id"] = self.tool_call_id

        if self.tool_calls:
            result["tool_calls"] = self.tool_calls

        return result

    def ancestor(self):
        node = self
        while node.parent:
            node = node.parent
        return node

    def history(self):
        node = self
        messages = [node.message()]

        while node.parent:
            node = node.parent
            messages.append(node.message())

        return messages[::-1]

    def __str__(self):
        return f"{self.role}: {self.content}"



"""Max tokens guard - enforces token limits on requests."""

from typing import TYPE_CHECKING, List

if TYPE_CHECKING:
    from src.chat import Chat
    from src.llm import LLM


def apply(chat: "Chat", llm: "LLM", config: dict) -> List[str]:
    """
    Enforce maximum token limit.

    Args:
        chat: Chat object
        llm: LLM object
        config: Guard configuration with 'max_tokens' value

    Returns:
        List of audit logs
    """
    max_tokens_limit = config.get("max_tokens")
    if max_tokens_limit is None:
        return []

    audit_logs = []

    # Check params in LLM
    # params is a dict
    current_max = llm.params.get("max_tokens")

    # Override if present and exceeds limit, or add if absent
    if current_max is None:
        llm.params["max_tokens"] = max_tokens_limit
        audit_logs.append(f"max_tokens: Enforced limit of {max_tokens_limit}")
    elif current_max > max_tokens_limit:
        llm.params["max_tokens"] = max_tokens_limit
        audit_logs.append(f"max_tokens: Enforced limit of {max_tokens_limit} (was {current_max})")

    return audit_logs



"""Guard application engine - applies guards to requests based on matching rules."""

import importlib
from typing import List, Tuple

from src import log
from src.chat import Chat
from src.guards import GuardBlockedError, GuardRule
from src.llm import LLM
from src.selection import match_filter

logger = log.setup_logger(__name__)


def apply_guards(chat: Chat, llm: LLM, guards: List[GuardRule]) -> Tuple[Chat, List[str]]:
    """
    Apply guards to a chat session in order.

    Args:
        chat: Chat object to process
        llm: LLM object containing parameters
        guards: List of guard rules to apply

    Returns:
        Tuple of (modified_chat, audit_logs)
    """
    audit_logs = []

    # Construct context for matching
    # This allows guards to match against model, params, messages, etc.
    match_context = {"model": llm.model, "messages": chat.history(), **llm.params}

    provider = getattr(llm, "provider", None)
    if provider is not None:
        match_context["provider"] = provider

    raw_payload = getattr(llm, "raw_payload", None)
    if raw_payload is not None:
        match_context["raw_payload"] = raw_payload

    for guard_idx, guard in enumerate(guards):
        # Check if guard matches the context
        try:
            matches = match_filter(match_context, guard.match)
        except Exception as e:
            logger.error(f"Error matching guard {guard_idx}: {e}")
            matches = False

        if not matches:
            continue

        logger.debug(f"Guard {guard_idx} matched, applying {len(guard.apply)} action(s)")

        # Apply each action in the guard
        for action in guard.apply:
            action_type = action.type
            action_config = action.config

            try:
                # Dynamically import the guard type module
                module_name = f"src.guard_types.{action_type}"
                guard_module = importlib.import_module(module_name)

                # Call the apply function
                # Signature: apply(chat, llm, config) -> List[str]
                action_logs = guard_module.apply(chat, llm, action_config)

                if action_logs:
                    audit_logs.extend(action_logs)

                logger.debug(f"Applied guard action '{action_type}'")

            except GuardBlockedError:
                raise
            except ModuleNotFoundError:
                logger.error(f"Guard type '{action_type}' not found")
            except AttributeError:
                logger.error(f"Guard type '{action_type}' does not have an 'apply' function")
            except Exception as e:
                if e.__class__.__name__ == "GuardBlockedError":
                    raise GuardBlockedError(str(e)) from e
                logger.error(f"Error applying guard action '{action_type}': {e}")

    return chat, audit_logs


def log_audit(audit_logs: List[str]):
    """
    Log audit entries.

    Args:
        audit_logs: List of audit log entries
    """
    for entry in audit_logs:
        logger.info(f"[AUDIT] {entry}")



"""LLM-driven input inspection guard."""

import json
import re
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Tuple

from src import log
from src.guards import GuardBlockedError

if TYPE_CHECKING:
    from src.chat import Chat
    from src.llm import LLM


logger = log.setup_logger(__name__)

DEFAULT_MAX_CHARS = 8000
MAX_ALLOWED_CHARS = 50000
VIOLATION_ACTIONS = {"block", "log"}
ERROR_ACTIONS = {"allow", "block"}
INSPECTION_SCHEMA = {
    "type": "object",
    "properties": {
        "decision": {"type": "string", "enum": ["allow", "block"]},
        "reason": {"type": "string"},
    },
    "required": ["decision"],
    "additionalProperties": False,
}


def apply(chat: "Chat", llm: "LLM", config: Dict[str, Any]) -> List[str]:
    """Inspect input text with an LLM and decide whether to allow or block."""
    prompt = str(config.get("prompt") or "").strip()
    if not prompt:
        logger.warning("llm_input_inspection: missing prompt, skipping")
        return []

    on_violation = _normalize_choice(config.get("on_violation"), VIOLATION_ACTIONS, "block")
    on_error = _normalize_choice(config.get("on_error"), ERROR_ACTIONS, "allow")
    max_chars = _normalize_max_chars(config.get("max_chars", DEFAULT_MAX_CHARS))
    inspector_model = config.get("inspector_model")

    inspected_text = _collect_inspected_text(chat, max_chars=max_chars)
    if not inspected_text:
        return []

    try:
        decision, reason = _inspect_with_llm(
            llm=llm,
            instructions=prompt,
            inspected_text=inspected_text,
            inspector_model=inspector_model,
        )
    except Exception as exc:
        error_msg = f"llm_input_inspection: inspection unavailable/error ({exc})"
        logger.warning(error_msg)
        if on_error == "block":
            raise GuardBlockedError("Request blocked: llm_input_inspection failed")
        return [f"{error_msg}; allowed by on_error=allow"]

    if decision == "block":
        reason_text = reason or "policy violation detected"
        if on_violation == "block":
            raise GuardBlockedError(f"Request blocked by llm_input_inspection: {reason_text}")
        return [f"llm_input_inspection: violation detected; on_violation=log; reason={reason_text}"]

    return []


def _normalize_choice(value: Any, allowed: set[str], default: str) -> str:
    normalized = str(value).strip().lower() if value is not None else default
    return normalized if normalized in allowed else default


def _normalize_max_chars(value: Any) -> int:
    try:
        parsed = int(value)
    except (TypeError, ValueError):
        parsed = DEFAULT_MAX_CHARS
    return max(1, min(parsed, MAX_ALLOWED_CHARS))


def _collect_inspected_text(chat: "Chat", max_chars: int) -> str:
    parts: List[str] = []

    for node in chat.plain():
        if getattr(node, "role", None) != "user":
            continue

        text = _extract_text(node.content)
        if text:
            parts.append(text)

    if not parts:
        return ""

    combined = "\n\n".join(parts).strip()
    if len(combined) <= max_chars:
        return combined

    return combined[:max_chars]


def _extract_text(content: Any) -> str:
    if isinstance(content, str):
        return content

    if isinstance(content, list):
        text_parts: List[str] = []
        for part in content:
            if isinstance(part, dict):
                if isinstance(part.get("text"), str):
                    text_parts.append(part["text"])
                elif part.get("type") == "text" and isinstance(part.get("content"), str):
                    text_parts.append(part["content"])
        return "\n".join(text_parts)

    if isinstance(content, dict):
        text = content.get("text")
        if isinstance(text, str):
            return text

    return ""


def _inspect_with_llm(
    llm: "LLM", instructions: str, inspected_text: str, inspector_model: Any
) -> Tuple[str, str]:
    model = inspector_model or getattr(llm, "model", None)

    inspect_completion = getattr(llm, "inspect_completion", None)
    if not callable(inspect_completion) or not model:
        raise RuntimeError("missing llm inspection capability in current context")

    system_prompt = (
        "You are an input safety inspector. "
        "Return a strict JSON object: "
        '{"decision":"allow|block","reason":"short reason"}. '
        "No extra text."
    )

    user_prompt = (
        f"Inspection policy:\n{instructions}\n\n"
        f"Candidate input:\n{inspected_text}\n\n"
        "Decide if this input should be blocked."
    )

    try:
        output_text = inspect_completion(
            system_prompt=system_prompt,
            user_prompt=user_prompt,
            model=model,
            response_format={
                "type": "json_schema",
                "json_schema": {
                    "name": "InputInspectionDecision",
                    "schema": INSPECTION_SCHEMA,
                },
            },
        )
    except Exception:
        output_text = inspect_completion(
            system_prompt=system_prompt,
            user_prompt=user_prompt,
            model=model,
            response_format=None,
        )

    return _parse_decision(output_text)


def _parse_decision(raw_text: str) -> Tuple[str, str]:
    if isinstance(raw_text, dict):
        decision, reason = _decision_from_object(raw_text)
        if decision is not None:
            return decision, reason
        raw_text = json.dumps(raw_text)

    text = str(raw_text or "").strip()
    if not text:
        raise ValueError("empty inspector output")

    parsed = _try_parse_json(text)
    if parsed is not None:
        decision, reason = _decision_from_object(parsed)
        if decision is not None:
            return decision, reason

    direct = _decision_from_text(text)
    if direct:
        return direct

    raise ValueError(f"unable to parse inspector decision from: {text[:200]}")


def _try_parse_json(text: str) -> Any:
    candidate = text
    if "```" in candidate:
        fence_match = re.search(r"```(?:json)?\s*(.*?)\s*```", candidate, re.DOTALL | re.IGNORECASE)
        if fence_match:
            candidate = fence_match.group(1).strip()

    try:
        return json.loads(candidate)
    except json.JSONDecodeError:
        pass

    start = candidate.find("{")
    end = candidate.rfind("}")
    if start != -1 and end != -1 and end > start:
        snippet = candidate[start : end + 1]
        try:
            return json.loads(snippet)
        except json.JSONDecodeError:
            return None

    return None


def _decision_from_object(data: Any) -> Tuple[Optional[str], str]:
    if isinstance(data, str):
        decision = _normalize_decision_word(data)
        if decision:
            return decision, ""
        return None, ""

    if isinstance(data, bool):
        return ("block", "policy violation") if data else ("allow", "")

    if not isinstance(data, dict):
        return None, ""

    decision_fields = ["decision", "verdict", "action", "result"]
    for field in decision_fields:
        value = data.get(field)
        decision = _normalize_decision_word(value)
        if decision:
            reason = str(data.get("reason") or data.get("explanation") or "").strip()
            return decision, reason

    if isinstance(data.get("block"), bool):
        return ("block", "policy violation") if data["block"] else ("allow", "")

    if isinstance(data.get("allow"), bool):
        return ("allow", "") if data["allow"] else ("block", "policy violation")

    return None, ""


def _normalize_decision_word(value: Any) -> str | None:
    if value is None:
        return None

    token = str(value).strip().lower()

    if token in {"allow", "allowed", "pass", "safe", "ok", "approve", "approved"}:
        return "allow"

    if token in {
        "block",
        "blocked",
        "deny",
        "denied",
        "reject",
        "rejected",
        "unsafe",
        "violation",
    }:
        return "block"

    return None


def _decision_from_text(text: str) -> Tuple[str, str] | None:
    normalized = text.strip().lower()

    decision_match = re.search(
        r"\b(decision|verdict|action|result)\b\s*[:=\-]\s*(allow|block)",
        normalized,
    )
    if decision_match:
        return decision_match.group(2), ""

    if re.fullmatch(r"\s*(allow|block)\s*[.!]?\s*", normalized):
        return normalized.strip(" .!"), ""

    has_block = bool(re.search(r"\b(block|deny|reject|unsafe|violation)\b", normalized))
    has_allow = bool(re.search(r"\b(allow|safe|ok|pass|approved?)\b", normalized))

    if has_block and not has_allow:
        return "block", "policy violation"
    if has_allow and not has_block:
        return "allow", ""

    return None



"""Keyword filter guard implementation."""

import re
from typing import TYPE_CHECKING, Any, List

from src.guards import GuardBlockedError

if TYPE_CHECKING:
    from src.chat import Chat
    from src.llm import LLM


def apply(chat: "Chat", llm: "LLM", config: dict) -> List[str]:
    """
    Apply keyword based filtering/blocking/logging.

    Args:
        chat: Chat object
        llm: LLM object
        config: Guard configuration

    Returns:
        List of audit logs
    """
    keywords = config.get("keywords", [])
    if not keywords:
        return []

    match_mode = config.get("match_mode", "any")  # "any" or "all"
    case_sensitive = config.get("case_sensitive", False)
    action = config.get("action", "block")  # "block", "sanitize", "log"
    replacement = config.get("replacement", "[REDACTED]")

    audit_logs = []

    flags = 0 if case_sensitive else re.IGNORECASE

    def get_text_content(content: Any) -> str:
        if isinstance(content, str):
            return content
        elif isinstance(content, list):
            text_parts = []
            for part in content:
                if isinstance(part, dict) and part.get("type") == "text":
                    text_parts.append(part.get("text", ""))
            return " ".join(text_parts)
        return str(content)

    def check_text(text: str, keyword: str) -> bool:
        pattern = re.escape(keyword)
        return bool(re.search(pattern, text, flags))

    # Identify which keywords are present in the payload
    present_keywords = set()

    for node in chat.plain():
        content = node.content
        text = get_text_content(content)
        for keyword in keywords:
            if check_text(text, keyword):
                present_keywords.add(keyword)

    # Determine if the guard should be triggered
    triggered = False
    trigger_reason = ""

    if match_mode == "any":
        if present_keywords:
            triggered = True
            trigger_reason = f"found keyword '{list(present_keywords)[0]}'"
    elif match_mode == "all":
        if set(keywords).issubset(present_keywords):
            triggered = True
            trigger_reason = "found all required keywords"

    if not triggered:
        return []

    # Execute Action
    if action == "block":
        if match_mode == "any":
            first_kw = list(present_keywords)[0]
            raise GuardBlockedError(f"Request blocked: found keyword '{first_kw}'")
        else:
            raise GuardBlockedError(f"Request blocked: {trigger_reason}")

    elif action == "log":
        audit_logs.append(f"Keyword filter triggered: {trigger_reason}")
        return audit_logs

    elif action == "sanitize":
        audit_logs.append(f"Sanitized keywords: {', '.join(present_keywords)}")

        def replace_in_text(text_val: str) -> str:
            new_text = text_val
            for kw in keywords:
                # If match_mode is 'any', we replace any found kw.
                # If match_mode is 'all', we replace all kw
                all_found = match_mode == "all" and set(keywords).issubset(present_keywords)
                if match_mode == "any" or all_found:
                    # Only replace if kw is in present_keywords?
                    # logic: sanitization should probably remove ALL keywords
                    # if trigger condition is met?
                    # Or only the ones found?
                    # If match_mode=all and we found all, we likely want to remove all.
                    if kw in present_keywords:
                        pattern = re.compile(re.escape(kw), flags)
                        new_text = pattern.sub(replacement, new_text)
            return new_text

        for node in chat.plain():
            content = node.content
            if isinstance(content, str):
                new_content = replace_in_text(content)
                if new_content != content:
                    node.content = new_content
            elif isinstance(content, list):
                for part in content:
                    if isinstance(part, dict) and part.get("type") == "text":
                        part["text"] = replace_in_text(part["text"])

    return audit_logs



from fastapi import Request
from starlette.middleware.base import BaseHTTPMiddleware

from src.state import request as request_state


class RequestStateMiddleware(BaseHTTPMiddleware):
    """
    Tracks current request in the context state.
    """

    async def dispatch(self, request: Request, call_next):
        request_state.set(request)
        response = await call_next(request)
        request_state.set(None)
        return response



import json
import shutil
from pathlib import Path

import typer
import uvicorn
import yaml

from src import config
from src import log as log_module
from src.guards import get_guards

# Setup logging
logger = log_module.setup_logger(__name__)

app = typer.Typer(no_args_is_help=False)


def _run_server():
    host = config.OPENGUARD_HOST.value
    port = config.OPENGUARD_PORT.value
    log_level = config.OPENGUARD_LOG_LEVEL.value.lower()

    logger.info(f"Starting OpenGuard on {host}:{port}")
    logger.info(f"Config file: {config.OPENGUARD_CONFIG.value}")

    # Load guards at startup
    guards = get_guards()
    logger.info(f"Loaded {len(guards)} guard rules")

    uvicorn.run("src.main:app", host=host, port=port, log_level=log_level, reload=False)


def install_opencode():
    """Configure OpenCode to use the local OpenGuard instance."""
    logger.info("Installing OpenGuard configuration for OpenCode...")

    # 1. Update ~/.local/share/opencode/auth.json
    auth_path = Path.home() / ".local/share/opencode/auth.json"

    if not auth_path.exists():
        logger.error(f"Auth file not found at {auth_path}")
        typer.echo(f"Error: Could not find OpenCode auth file at {auth_path}")
        typer.echo("Please ensure OpenCode is installed and has run at least once.")
        return

    try:
        # Create backup
        backup_path = auth_path.with_suffix(".json.bak")
        shutil.copy2(auth_path, backup_path)
        logger.info(f"Backed up auth.json to {backup_path}")

        with open(auth_path, "r") as f:
            auth_data = json.load(f)

        # Migrating existing credentials to OpenGuard config
        providers_list = []

        # Check for Base URLs in opencode.json
        local_opencode_path = Path.cwd() / "opencode.json"

        opencode_providers_conf = {}
        if local_opencode_path.exists():
            try:
                with open(local_opencode_path, "r") as f:
                    local_data = json.load(f)
                    opencode_providers_conf = local_data.get("provider", {})
            except Exception:
                pass

        for provider_name, creds in auth_data.items():
            if provider_name == "openguard":
                continue

            if creds.get("type") == "oauth":
                logger.warning(f"Provider '{provider_name}' uses OAuth. Skipping migration.")
                continue

            if creds.get("type") == "api" and "key" in creds:
                provider_key = creds["key"]
                provider_url = ""

                # Check known URL for provider in opencode.json
                p_conf = opencode_providers_conf.get(provider_name)
                if p_conf:
                    provider_url = (
                        p_conf.get("baseURL") or p_conf.get("options", {}).get("baseURL") or ""
                    )

                # Determine type
                ptype = "openai"
                if "anthropic" in provider_name.lower():
                    ptype = "anthropic"

                providers_list.append({"type": ptype, "key": provider_key, "url": provider_url})

        if providers_list:
            # Write to OpenGuard config
            og_config_dir = Path.home() / ".config" / "openguard"
            og_config_dir.mkdir(parents=True, exist_ok=True)
            og_config_file = og_config_dir / "config.yaml"

            og_data = {}
            if og_config_file.exists():
                try:
                    with open(og_config_file, "r") as f:
                        og_data = yaml.safe_load(f) or {}
                except Exception:
                    pass

            og_data["providers"] = providers_list

            with open(og_config_file, "w") as f:
                yaml.safe_dump(og_data, f)

            logger.info(f"Migrated credentials to {og_config_file}")

        # Add openguard credential
        auth_data["openguard"] = {"type": "api", "key": "sk-openguard"}

        with open(auth_path, "w") as f:
            json.dump(auth_data, f, indent=2)

        logger.info("Added 'openguard' credentials to auth.json")

    except Exception as e:
        logger.error(f"Failed to update auth.json: {e}")
        typer.echo(f"Error updating auth.json: {e}")
        return

    # 2. Update opencode.json in current directory
    config_path = Path.cwd() / "opencode.json"
    config_data = {}

    if config_path.exists():
        try:
            with open(config_path, "r") as f:
                config_data = json.load(f)
        except json.JSONDecodeError:
            logger.warning(
                f"Could not parse existing {config_path} (likely JSONC/comments). Backup and reset."
            )
            typer.echo(
                f"Warning: Could not parse {config_path} (might be JSONC). Creating fresh config."
            )
            shutil.copy2(config_path, config_path.with_suffix(".json.bak"))
            config_data = {}

    # Configure provider
    if "provider" not in config_data:
        config_data["provider"] = {}

    config_data["provider"]["openguard"] = {
        "npm": "@ai-sdk/openai-compatible",
        "name": "OpenGuard",
        "options": {
            "baseURL": f"http://{config.OPENGUARD_HOST.value}:{config.OPENGUARD_PORT.value}/v1"
        },
        "models": {
            "gpt-4o": {"name": "GPT-4o (Guarded)"},
            "claude-3-5-sonnet": {"name": "Claude 3.5 Sonnet (Guarded)"},
        },
    }

    # Set default model to verify installation
    # Overwrite model if it's set, or set if missing. Prompt suggests overwrite is fine.
    config_data["model"] = "openguard:gpt-4o"

    try:
        with open(config_path, "w") as f:
            json.dump(config_data, f, indent=2)
        logger.info(f"Updated {config_path} with OpenGuard provider")
    except Exception as e:
        logger.error(f"Failed to update opencode.json: {e}")
        typer.echo(f"Error updating opencode.json: {e}")
        return

    typer.echo("✅ OpenGuard installed for OpenCode successfully!")
    typer.echo("Run 'openguard serve' (or 'make dev') to start the proxy.")


@app.command()
def serve():
    """Start the OpenGuard server."""
    _run_server()


@app.command()
def install(service: str):
    """Install OpenGuard for a specific service."""
    if service.lower() == "opencode":
        install_opencode()
    else:
        typer.echo(f"Service '{service}' not supported yet.")


@app.callback(invoke_without_command=True)
def main(ctx: typer.Context):
    """
    OpenGuard CLI - Guarding proxy for AI.
    """
    if ctx.invoked_subcommand is None:
        _run_server()


if __name__ == "__main__":
    app()



from typing import TYPE_CHECKING, Any, Dict, Mapping, Optional

import httpx
from asyncache import cached
from cachetools import TTLCache
from fastapi import HTTPException

from . import config, log

if TYPE_CHECKING:
    from . import llm

logger = log.setup_logger(__name__)

MODEL_TO_BACKEND: Dict[str, Dict[str, str]] = {"openai": {}, "anthropic": {}}

# TODO: Add mods (modules) support when adoption is complete
# import mods


@cached(TTLCache(1024, 60))
async def list_downstream(provider: str = "openai"):
    logger.debug(f"Listing downstream models for provider '{provider}'")

    all_models = []

    for backend in get_provider_backends(provider):
        url = backend["url"]
        key = backend["key"]
        try:
            endpoint = build_provider_endpoint(url, provider, "/models")
            headers = {
                "Content-Type": "application/json",
            }
            if key:
                if provider == "anthropic":
                    headers["x-api-key"] = key
                else:
                    headers["Authorization"] = f"Bearer {key}"

            logger.debug(f"Fetching models from '{endpoint}'")

            async with httpx.AsyncClient() as client:
                response = await client.get(endpoint, headers=headers)
                response.raise_for_status()
                json = response.json()
                models = json.get("data", [])

                logger.debug(f"Found {len(models)} models at '{endpoint}'")
                all_models.extend(models)

                for model in models:
                    model_id = model.get("id")
                    if model_id:
                        MODEL_TO_BACKEND.setdefault(provider, {})[model_id] = url

        except Exception as e:
            logger.error(f"Failed to fetch models from {endpoint}: {e}")

    return all_models


def _as_list(value):
    if isinstance(value, str):
        return [value]
    return list(value or [])


def get_provider_backends(provider: str):
    provider_normalized = provider.lower()
    if provider_normalized == "anthropic":
        urls = _as_list(config.OPENGUARD_ANTHROPIC_URLS.value)
        keys = _as_list(config.OPENGUARD_ANTHROPIC_KEYS.value)
    else:
        urls = _as_list(config.OPENGUARD_OPENAI_URLS.value)
        keys = _as_list(config.OPENGUARD_OPENAI_KEYS.value)

    if len(keys) < len(urls):
        keys.extend([""] * (len(urls) - len(keys)))

    backends = []
    for index, url in enumerate(urls):
        if not url:
            continue
        backends.append({"provider": provider_normalized, "url": url, "key": keys[index]})

    return backends


def get_internal_api_keys():
    api_key = config.OPENGUARD_API_KEY.value
    api_keys_list = _as_list(config.OPENGUARD_API_KEYS.value)

    keys = []
    if api_key:
        keys.append(api_key)
    keys.extend(k for k in api_keys_list if k)
    return keys


def extract_api_key(headers: Optional[Mapping[str, str]]):
    if not headers:
        return ""

    x_api_key = headers.get("x-api-key") or headers.get("X-API-Key")
    if x_api_key:
        return x_api_key

    auth_header = headers.get("authorization") or headers.get("Authorization") or ""
    if auth_header.startswith("Bearer "):
        return auth_header[7:]

    return ""


def build_provider_endpoint(base_url: str, provider: str, path: str):
    base = base_url.rstrip("/")
    normalized_path = path if path.startswith("/") else f"/{path}"

    if provider.lower() == "anthropic":
        if normalized_path.startswith("/v1/"):
            return f"{base}{normalized_path}"
        if base.endswith("/v1"):
            return f"{base}{normalized_path}"
        return f"{base}/v1{normalized_path}"

    return f"{base}{normalized_path}"


def _resolve_backend_for_model(provider: str, model: Optional[str]):
    if model:
        cached = MODEL_TO_BACKEND.get(provider, {}).get(model)
        if cached:
            for backend in get_provider_backends(provider):
                if backend["url"] == cached:
                    return backend

    backends = get_provider_backends(provider)
    if len(backends) == 1:
        return backends[0]

    return None


def resolve_provider_route(provider: str, endpoint_path: str, body: Optional[Dict], headers):
    provider_normalized = provider.lower()
    backends = get_provider_backends(provider_normalized)
    if not backends:
        raise ValueError(f"No downstream backends configured for provider '{provider_normalized}'")

    incoming_key = extract_api_key(headers)
    internal_keys = get_internal_api_keys()

    if incoming_key and incoming_key in internal_keys:
        internal_backend = _resolve_backend_for_model(
            provider_normalized, (body or {}).get("model")
        )
        if internal_backend is None:
            internal_backend = backends[0]

        return {
            **internal_backend,
            "mode": "internal",
            "endpoint": build_provider_endpoint(
                internal_backend["url"], provider_normalized, endpoint_path
            ),
        }

    if incoming_key:
        for backend in backends:
            if backend["key"] and backend["key"] == incoming_key:
                return {
                    **backend,
                    "mode": "direct",
                    "endpoint": build_provider_endpoint(
                        backend["url"], provider_normalized, endpoint_path
                    ),
                }

    backend = backends[0]
    return {
        **backend,
        "mode": "endpoint",
        "endpoint": build_provider_endpoint(backend["url"], provider_normalized, endpoint_path),
    }


def get_proxy_model(module, model: dict) -> Dict:
    return {
        **model,
        "id": f"{module.ID_PREFIX}-{model['id']}",
        "name": f"{module.ID_PREFIX} {model['id']}",
    }


def resolve_proxy_model(model_id: str) -> str:
    # Logic disabled until mods are ported
    # parts = model_id.split("-")
    # if parts[0] in mods.registry:
    #   return "-".join(parts[1:])
    return model_id


def resolve_proxy_module(model_id: str) -> Optional[str]:
    # Logic disabled until mods are ported
    # parts = model_id.split("-")
    # if parts[0] in mods.registry:
    #   return parts[0]
    return None


def resolve_request_config(body: Dict) -> Dict:
    model = body.get("model")
    messages = body.get("messages")
    params = {k: v for k, v in body.items() if k not in ["model", "messages"]}

    if not model:
        raise ValueError("Unable to proxy request without a model specifier")

    proxy_model = resolve_proxy_model(model)
    proxy_module = resolve_proxy_module(model)
    proxy_backend = MODEL_TO_BACKEND.get("openai", {}).get(proxy_model)

    logger.debug(
        f"Resolved proxy model: {proxy_model}, proxy module: {proxy_module}, "
        f"proxy backend: {proxy_backend}"
    )

    if not proxy_backend:
        # If not found in cache, it might be that we haven't fetched models yet or it's invalid.
        # For now, if we can't find backend, we might fail or default to first one.
        # Current logic strictly requires backend map.

        # Try to re-populate cache if empty or missing?
        # list_downstream is cached.

        # Fallback: check if we can guess backend from config if only 1 exists
        urls = _as_list(config.OPENGUARD_OPENAI_URLS.value)

        if len(urls) == 1:
            proxy_backend = urls[0]
        else:
            # Check if we should raise HTTPException or ValueError based on contract
            # Standard behavior here seems to ideally be 400/404
            raise HTTPException(
                status_code=404,
                detail=f"Unknown model: '{model}'",
            )

    # Find key for backend
    urls = _as_list(config.OPENGUARD_OPENAI_URLS.value)
    keys = _as_list(config.OPENGUARD_OPENAI_KEYS.value)

    proxy_key = ""
    if proxy_backend in urls:
        idx = urls.index(proxy_backend)
        if idx < len(keys):
            proxy_key = keys[idx]

    proxy_config: Dict[str, Any] = {
        "url": proxy_backend,
        "headers": {
            "Content-Type": "application/json",
        },
        "model": proxy_model,
        "params": params,
        "messages": messages,
        "module": proxy_module,
    }

    if proxy_key:
        proxy_config["headers"]["Authorization"] = f"Bearer {proxy_key}"

    return proxy_config


def is_title_generation_task(llm: "llm.LLM"):
    # TODO: Better way to identify?
    return llm.chat.has_substring("3-5 word title")


DIRECT_TASK_PROMPTS = [
    # Open WebUI prompts related to system tasks
    "Generate a concise, 3-5 word title",
    "Based on the chat history, determine whether a search is necessary",
    "Generate 1-3 broad tags categorizing",
    "You are an autocompletion system. Continue the text in `<text>` based on the "
    "**completion type**",
    "determine the necessity of generating search queries",
    # Custom for the test
    "[{DIRECT}]",
]


def is_direct_task(llm: "llm.LLM"):
    return any(llm.chat.has_substring(prompt) for prompt in DIRECT_TASK_PROMPTS)



import re

from . import config

status_formatters = {
    "md:codeblock": "\n```boost\n{status}\n```\n",
    "md:h1": "\n\n# {status}\n\n",
    "md:h2": "\n\n## {status}\n\n",
    "md:h3": "\n\n### {status}\n\n",
    "plain": "\n\n{status}\n\n",
    "none": "",
}


def format_status(status: str):
    desired_format = config.STATUS_STYLE.value

    if desired_format not in status_formatters:
        desired_format = "md:codeblock"

    return status_formatters[desired_format].format(status=status)


def format_artifact(artifact: str):
    return f"\n```html\n{artifact}\n```\n"


def remove_html_code_blocks(text_content):
    """
    Removes all HTML markdown code blocks (```html...```) from a string.

    Args:
        text_content (str): The string containing potential HTML code blocks.

    Returns:
        str: The string with HTML code blocks removed.
    """
    regex_pattern = r"```html\n(.*?)\n```"
    cleaned_text = re.sub(regex_pattern, "", text_content, flags=re.DOTALL)
    return cleaned_text



import uuid
from contextvars import ContextVar

from fastapi import Request
from starlette.middleware.base import BaseHTTPMiddleware

request_id_var: ContextVar[str] = ContextVar("request_id", default="")


class RequestIDMiddleware(BaseHTTPMiddleware):
    async def dispatch(self, request: Request, call_next):
        default_request_id = str(uuid.uuid4())[:8]
        request_id = request.headers.get("X-Request-ID", default_request_id)
        request_id_var.set(request_id)
        response = await call_next(request)
        response.headers["X-Request-ID"] = request_id

        return response



import inspect
import re
from typing import Any, List, Optional, Tuple, Union

DottyPath = Union[str, List, Tuple]


def is_int(value: Any) -> bool:
    """Check if the value is an integer or can be converted to an integer."""
    if isinstance(value, int):
        return True
    if isinstance(value, str):
        return value.isdigit() or (value.startswith("-") and value[1:].isdigit())
    return False


def path_to_string(path: DottyPath) -> str:
    """Serialise a path string or list/tuple into a dot notation string."""
    if isinstance(path, (list, tuple)):
        return ".".join(str(part) for part in path)

    if not path:
        return ""

    parts = parse_path(path)
    return ".".join(str(part) for part in parts)


def parse_path(path: DottyPath) -> List[Union[str, int]]:
    """Parse a path string or list/tuple into a list of path segments."""
    if isinstance(path, (list, tuple)):
        return list(path)

    if not path:
        return []

    result: List[Union[str, int]]

    # Handle bracket notation like 'foo[0][bar]'
    if "[" in path:
        parts = re.findall(r"([^\.\[\]]+|\[\d+\]|\[[^\[\]]+\])", path)
        result = []
        for part in parts:
            if part.startswith("[") and part.endswith("]"):
                content = part[1:-1]
                if is_int(content):
                    result.append(int(content))
                else:
                    result.append(content)
            else:
                result.append(part)
        return result

    # Handle dot notation like 'foo.bar' or 'foo.0.bar'
    result = []
    for part in path.split("."):
        if is_int(part):
            result.append(int(part))
        else:
            result.append(part)
    return result


def _detect_callable_type(fn: Any) -> str:
    """
    Detect the type of callable to determine how to invoke it properly.

    Returns:
      String describing the callable type: "bound_method", "unbound_method",
      "class_constructor", "function", or "callable"
    """
    # Check if it's a bound method (has __self__ that is not None)
    if hasattr(fn, "__self__") and fn.__self__ is not None:
        return "bound_method"

    # Check for method objects (bound methods)
    if inspect.ismethod(fn):
        return "bound_method"

    # Check if it's a class constructor
    if inspect.isclass(fn):
        return "class_constructor"

    # Check if it's a regular function or builtin
    if inspect.isfunction(fn) or inspect.isbuiltin(fn):
        # Only consider it an unbound method if:
        # 1. It's a regular function (not builtin)
        # 2. It has a __qualname__ with a dot (indicating it's defined in a class)
        # 3. It doesn't have __self__ (which would make it bound)
        # 4. The first parameter is named 'self' or 'cls' (indicating it's an instance/class method)
        if (
            inspect.isfunction(fn)
            and hasattr(fn, "__qualname__")
            and "." in fn.__qualname__
            and not hasattr(fn, "__self__")
        ):
            # Check if first parameter is 'self' or 'cls'
            try:
                sig = inspect.signature(fn)
                params = list(sig.parameters.keys())
                if params and params[0] in ("self", "cls"):
                    return "unbound_method"
            except (ValueError, TypeError):
                pass

        return "function"

    # Fallback for other callables
    return "callable"


def call(
    obj: Any, path: DottyPath, args: Optional[List[Any]] = None, kwargs: Optional[dict] = None
) -> Any:
    """
    Call a function or method at the given path with the provided arguments.
    Automatically handles different types of callables including bound methods,
    unbound methods, class constructors, and regular functions.

    For unbound methods, automatically infers the instance (self) from the parent
    object in the path. For example, calling obj.instance.method will automatically
    pass obj.instance as the first argument to the unbound method.

    Args:
      obj: The object to query for the callable
      path: The path to the callable (string, list, or tuple)
      args: List of positional arguments to pass to the callable
      kwargs: Dictionary of keyword arguments to pass to the callable

    Returns:
      The result of calling the function/method

    Raises:
      ValueError: If the path doesn't resolve to a callable
      TypeError: If an unbound method is called without an instance and no parent can be inferred
    """
    if args is None:
        args = []
    if kwargs is None:
        kwargs = {}

    maybe_fn = get(obj, path)

    if not callable(maybe_fn):
        raise ValueError(f"Path '{path}' does not resolve to a callable")

    callable_type = _detect_callable_type(maybe_fn)

    # Call based on detected type
    if callable_type == "bound_method":
        return maybe_fn(*args, **kwargs)
    elif callable_type == "unbound_method":
        # Auto-infer self for unbound methods by using parent object in path
        parts = parse_path(path)
        if len(parts) > 1:
            # Get the parent object (everything except the last part of the path)
            parent_path = parts[:-1]
            parent_obj = get(obj, parent_path)
            # Only use parent as self if it's an instance (not a class)
            if parent_obj is not None and not inspect.isclass(parent_obj):
                # Prepend parent object as self to the args
                args = [parent_obj] + list(args)

        if not args:
            raise TypeError("Unbound method requires instance as first argument")
        return maybe_fn(*args, **kwargs)
    elif callable_type == "class_constructor":
        return maybe_fn(*args, **kwargs)
    elif callable_type == "function":
        return maybe_fn(*args, **kwargs)
    else:
        return maybe_fn(*args, **kwargs)


def get(obj: Any, path: DottyPath, default: Any = None) -> Any:
    """
    Gets the value at path of object. If the resolved value is undefined,
    the default value is returned. Works with nested dictionaries, lists,
    class attributes, and properties.

    Args:
        obj: The object to query
        path: The path of the property to get (string, list, or tuple)
        default: The value returned for undefined resolved values

    Returns:
        The resolved value or default if the path doesn't exist
    """
    if obj is None:
        return default

    parts = parse_path(path)

    try:
        result = obj
        for part in parts:
            if result is None:
                return default

            if isinstance(result, dict) and part in result:
                result = result[part]
            elif (
                isinstance(result, (list, tuple))
                and isinstance(part, int)
                and -len(result) <= part < len(result)
            ):
                result = result[part]
            elif hasattr(result, part):
                result = getattr(result, part)
            else:
                try:
                    # Try to access as dictionary/list even if not isinstance
                    result = result[part]
                except (KeyError, TypeError, IndexError):
                    return default

        if result is None:
            return default

        return result
    except (KeyError, AttributeError, IndexError, TypeError):
        return default


def set(obj: Any, path: DottyPath, value: Any) -> Any:
    """
    Sets the value at path of object. If a portion of path doesn't exist, it's created.
    Works with nested dictionaries, lists, and class attributes.

    Args:
        obj: The object to modify
        path: The path of the property to set (string, list, or tuple)
        value: The value to set

    Returns:
        The modified object
    """
    if obj is None:
        # If the root is None, create a new dict as the root
        obj = {}

    parts = parse_path(path)
    if not parts:
        return value

    root = obj
    for i, part in enumerate(parts[:-1]):
        last_part = i == len(parts) - 2
        next_part = parts[i + 1]

        # Handle dictionary access
        if isinstance(root, dict):
            if part not in root:
                root[part] = [] if isinstance(next_part, int) else {}
            root = root[part]
        # Handle list access
        elif isinstance(root, list) and isinstance(part, int):
            # Extend list if needed
            while len(root) <= part:
                root.append(None)
            if root[part] is None:
                root[part] = [] if isinstance(next_part, int) else {}
            root = root[part]
        # Handle object attributes
        elif hasattr(root, part):
            current = getattr(root, part)
            if current is None and last_part:
                setattr(root, part, [] if isinstance(next_part, int) else {})
            root = getattr(root, part)
        # Try to set attribute if it doesn't exist
        else:
            try:
                setattr(root, part, [] if isinstance(next_part, int) else {})  # type: ignore[arg-type]
                root = getattr(root, part)  # type: ignore[arg-type]
            except (AttributeError, TypeError):
                # If we can't set the attribute, fall back to treating as dict
                try:
                    root[part] = [] if isinstance(next_part, int) else {}
                    root = root[part]
                except (TypeError, KeyError):
                    # If all else fails, we can't navigate further
                    return obj

    # Set the final value
    last_part = parts[-1]
    if isinstance(root, dict):
        root[last_part] = value
    elif isinstance(root, list) and isinstance(last_part, int):
        # Extend list if needed
        while len(root) <= last_part:
            root.append(None)
        root[last_part] = value
    elif hasattr(root, last_part):
        setattr(root, last_part, value)
    else:
        try:
            # Try to set as attribute
            setattr(root, last_part, value)
        except (AttributeError, TypeError):
            # Fall back to treating as dict/list
            try:
                root[last_part] = value
            except (TypeError, IndexError):
                pass

    return obj


def has(obj: Any, path: DottyPath) -> bool:
    """
    Checks if path is a direct property of object.
    Works with nested dictionaries, lists, and class attributes.

    Args:
        obj: The object to query
        path: The path to check (string, list, or tuple)

    Returns:
        True if path exists, False otherwise
    """
    if obj is None:
        return False

    parts = parse_path(path)
    if not parts:
        return False

    try:
        result = obj
        for part in parts:
            if result is None:
                return False

            if isinstance(result, dict):
                if part not in result:
                    return False
                result = result[part]
            elif isinstance(result, (list, tuple)):
                if not isinstance(part, int) or part < -len(result) or part >= len(result):
                    return False
                result = result[part]
            elif hasattr(result, part):
                result = getattr(result, part)
            else:
                try:
                    # Try to access as dictionary/list even if not isinstance
                    result = result[part]
                except (KeyError, TypeError, IndexError):
                    return False
        return True
    except (KeyError, AttributeError, IndexError, TypeError):
        return False


def update(obj: Any, path: DottyPath, value: Any) -> Any:
    """
    Updates the dictionaty at the path of object. If the leaf value is not
    a dictionary - an error is raised. If a portion of path doesn't exist,
    it's created.

    Args:
        obj: The object to modify
        path: The path of the property to update (string, list, or tuple)
        value: The value to update
    Returns:
        The modified object
    """

    existing = get(obj, path, {})
    if not isinstance(existing, dict):
        raise TypeError(f"Cannot update non-dictionary value at path: {path}")

    existing.update(value)
    set(obj, path, existing)
    return obj


def interpolate(string: str, context: dict) -> str:
    """
    Interpolates the string using the context dictionary.
    Supports dot notation for nested values.

    Args:
        string: The string to interpolate
        context: The context dictionary for interpolation

    Returns:
        The interpolated string
    """
    pattern = re.compile(r"\{([^\}]+)\}")

    def replacer(match):
        path = match.group(1)
        return str(get(context, path, ""))

    return pattern.sub(replacer, string)


def prefix(obj: Any, path: DottyPath, prefix: str) -> Any:
    set(obj, path, prefix + get(obj, path, ""))
    return obj


def pop(obj: Any, path: DottyPath, default: Any = None) -> Any:
    """
    Pops the value at path of object. If the resolved value is undefined,
    the default value is returned. Works with nested dictionaries, lists,
    class attributes, and properties.

    Args:
        obj: The object to query
        path: The path of the property to pop (string, list, or tuple)
        default: The value returned for undefined resolved values

    Returns:
        The popped value or the default value
    """
    value = get(obj, path, default)
    set(obj, path, None)
    return value


def increment(obj: Any, path: DottyPath, amount: int = 1) -> Any:
    current = get(obj, path, 0)
    if not isinstance(current, (int, float)):
        current = 0
    set(obj, path, current + amount)
    return obj


def transform(obj: Any, path: DottyPath, fn) -> Any:
    current = get(obj, path)
    new_value = fn(current)
    set(obj, path, new_value)
    return new_value


def delete(obj: Any, path: DottyPath) -> Any:
    """
    Deletes the value at path of object.
    Works with nested dictionaries, lists, and class attributes.

    Args:
      obj: The object to modify
      path: The path of the property to delete (string, list, or tuple)
    """

    parts = parse_path(path)
    if not parts:
        return obj

    # Resolve parent using existing get() logic to keep behavior consistent
    parent_path = parts[:-1]
    root = get(obj, parent_path)
    if root is None:
        return obj

    last_part = parts[-1]
    if isinstance(root, dict) and last_part in root:
        del root[last_part]
    elif (
        isinstance(root, list)
        and isinstance(last_part, int)
        and -len(root) <= last_part < len(root)
    ):
        del root[last_part]
    elif hasattr(root, last_part):  # type: ignore[arg-type]
        delattr(root, last_part)  # type: ignore[arg-type]
    else:
        try:
            del root[last_part]
        except (KeyError, TypeError, IndexError):
            pass

    return obj



import asyncio
from collections import defaultdict
from collections.abc import Callable

from src import log

logger = log.setup_logger(__name__)


class AsyncEventEmitter:
    def __init__(self):
        self._listeners = defaultdict(list)
        self._listeners_once = defaultdict(list)
        self._lock = asyncio.Lock()

    async def on(self, event_name: str, listener: Callable):
        async with self._lock:
            self._listeners[event_name].append(listener)
        logger.debug(f"ON: {listener.__name__} for '{event_name}'")  # type: ignore

    async def once(self, event_name: str, listener: Callable):
        if not asyncio.iscoroutinefunction(listener):
            raise TypeError("Listener must be async")
        async with self._lock:
            self._listeners_once[event_name].append(listener)
        logger.debug(f"ONCE: {listener.__name__} for '{event_name}'")  # type: ignore

    async def off(self, event_name: str, listener: Callable):
        async with self._lock:
            removed = False
            if listener in self._listeners[event_name]:
                self._listeners[event_name].remove(listener)
                removed = True
            elif listener in self._listeners_once[event_name]:
                self._listeners_once[event_name].remove(listener)
                removed = True

            if removed:
                logger.debug(f"OFF: {listener.__name__} from '{event_name}'")  # type: ignore
            else:
                logger.warning(f"OFF_NOT_FOUND: {listener.__name__} for '{event_name}'")  # type: ignore

    async def remove_all_listeners(self, event_name: str | None = None):
        async with self._lock:
            if event_name:
                if event_name in self._listeners:
                    del self._listeners[event_name]
                if event_name in self._listeners_once:
                    del self._listeners_once[event_name]
                logger.debug(f"REMOVED_ALL: Listeners for '{event_name}'")
            else:
                self._listeners.clear()
                self._listeners_once.clear()
                logger.debug("REMOVED_ALL: All listeners for all events")

    async def emit(self, event_name: str, *args, **kwargs):
        listeners_to_call = []
        once_listeners_to_call = []

        async with self._lock:
            if event_name in self._listeners:
                listeners_to_call.extend(self._listeners[event_name])
            if event_name in self._listeners_once:
                once_listeners_to_call.extend(self._listeners_once[event_name])
                self._listeners_once[event_name].clear()

        if not listeners_to_call and not once_listeners_to_call:
            logger.debug(f"EMIT_NO_LISTENERS: '{event_name}'")
            return

        logger.info(
            f"EMIT: '{event_name}' "
            f"(Listeners: {len(listeners_to_call) + len(once_listeners_to_call)})"
        )

        all_listeners = listeners_to_call + once_listeners_to_call
        tasks = [
            self._call_listener(listener, event_name, *args, **kwargs) for listener in all_listeners
        ]
        await asyncio.gather(*tasks, return_exceptions=False)

    async def _call_listener(self, listener: Callable, event_name: str, *args, **kwargs):
        try:
            await listener(*args, **kwargs)
        except Exception as e:
            logger.error(
                f"LISTENER_ERROR: {listener.__name__} on '{event_name}': {e}",
                exc_info=False,  # type: ignore
            )



import logging

try:
    from src.middleware import request_id

    HAS_MIDDLEWARE = True
except ImportError:
    HAS_MIDDLEWARE = False

# Common ANSI color codes for reuse
COLORS = {
    "black": "\033[30m",
    "red": "\033[31m",
    "green": "\033[32m",
    "yellow": "\033[33m",
    "blue": "\033[34m",
    "magenta": "\033[35m",
    "cyan": "\033[36m",
    "white": "\033[37m",
    "gray": "\033[90m",
    "reset": "\033[0m",
    "bold": "\033[1m",
    "underline": "\033[4m",
}


class BoostFormatter(logging.Formatter):
    COLORS = {
        "DEBUG": COLORS["gray"],
        "INFO": COLORS["blue"],
        "WARNING": COLORS["yellow"],
        "ERROR": COLORS["red"],
        "CRITICAL": "\033[41m",  # Keep red background as it's not in COLORS map
    }

    LEVEL_LABELS = {
        "DEBUG": "⏺ ",
        "INFO": "⏺ ",
        "WARNING": "⏺ // ",
        "ERROR": "⏺ !! ",
        "CRITICAL": "⏺ !! ",
    }

    def format(self, record):
        color = self.COLORS.get(record.levelname, COLORS["reset"])
        label = self.LEVEL_LABELS.get(record.levelname, record.levelname)
        reset = COLORS["reset"]
        # For debug messages, make the entire message gray
        if record.levelname == "DEBUG":
            record.msg = f"{color}{record.msg}{reset}"
        record.levelname = f"{color}{label}{reset}"

        # Add request ID to log messages if middleware is available
        if HAS_MIDDLEWARE:
            req_id = request_id.request_id_var.get()
            if req_id:
                record.levelname = f"{color}{req_id}{reset} {record.levelname}"

        return super().format(record)


formatter = BoostFormatter("%(levelname)s%(name)s %(message)s", datefmt="%H:%M:%S")


def setup_logger(name):
    logger = logging.getLogger(name)
    if not logger.handlers:
        logger.setLevel(logging.DEBUG)
        handler = logging.StreamHandler()
        handler.set_name(name)
        handler.setFormatter(formatter)
        logger.addHandler(handler)
        logger.propagate = False
    return logger



import json
import os
from pathlib import Path
from typing import Any, Dict, Generic, List, Optional, Type, TypeVar, Union

import yaml

# ----------------- Persistent Configuration -----------------

PERSISTENT_CONFIG: Dict[str, Any] = {}


def _load_persistent_config():
    global PERSISTENT_CONFIG
    config_dir = Path.home() / ".config" / "openguard"
    config_file = config_dir / "config.yaml"
    json_config_file = config_dir / "config.json"

    data = {}
    if config_file.exists():
        try:
            with open(config_file, "r") as f:
                data = yaml.safe_load(f) or {}
        except Exception:
            pass
    elif json_config_file.exists():
        try:
            with open(json_config_file, "r") as f:
                data = json.load(f)
        except Exception:
            pass

    if data:
        PERSISTENT_CONFIG.update(data)


_load_persistent_config()

T = TypeVar("T")


class ConfigDict(Dict[str, Union[str, int, float, bool]]):
    @classmethod
    def from_string(cls, value: str) -> "ConfigDict":
        result = cls()
        if not value:
            return result

        # Handle if value is somehow passed as a dict (defensive)
        if isinstance(value, dict):
            return cls(value)

        pairs = value.split(",")
        for pair in pairs:
            if "=" not in pair:
                continue
            key, val = pair.split("=", 1)
            key = key.strip()
            val = val.strip()
            # Try to parse the value as int, float, or bool
            if val.lower() == "true":
                result[key] = True
            elif val.lower() == "false":
                result[key] = False
            else:
                try:
                    result[key] = int(val)
                except ValueError:
                    try:
                        result[key] = float(val)
                    except ValueError:
                        result[key] = val
        return result


class StrList(List[str]):
    @classmethod
    def from_string(cls, value: str) -> "StrList":
        return (
            cls(item.strip() for item in value.split(";") if item.strip())
            if value.strip()
            else cls()
        )


class FloatList(List[float]):
    @classmethod
    def from_string(cls, value: str) -> "FloatList":
        return (
            cls(float(item.strip()) for item in value.split(";") if item.strip())
            if value.strip()
            else cls()
        )


class BoolList(List[bool]):
    @classmethod
    def from_string(cls, value: str) -> "BoolList":
        return (
            cls(item.strip().lower() == "true" for item in value.split(";") if item.strip())
            if value.strip()
            else cls()
        )


class IntList(List[int]):
    @classmethod
    def from_string(cls, value: str) -> "IntList":
        return (
            cls(int(item.strip()) for item in value.split(";") if item.strip())
            if value.strip()
            else cls()
        )


class Config(Generic[T]):
    name: str
    type: Type[T]
    default: str
    description: Optional[str]
    __value__: T

    def __init__(self, name: str, type: Type[T], default: str, description: Optional[str] = None):
        self.name = name
        self.type = type
        self.default = default
        self.description = description
        self.__value__ = self.resolve_value()  # type: ignore

    @property
    def value(self) -> T:
        return self.__value__

    def resolve_value(self) -> Union[T, List[T]]:
        if "*" in self.name:
            return self._resolve_wildcard()
        else:
            return self._resolve_single()

    def _resolve_single(self) -> T:
        # 1. Check persistent config (mapping OPENGUARD_XYZ -> xyz)
        config_key = self.name.replace("OPENGUARD_", "").lower()
        if config_key in PERSISTENT_CONFIG:
            val = PERSISTENT_CONFIG[config_key]
            # For complex types like StrList, ConfigDict, handle direct assignment from YAML types
            if issubclass(self.type, (ConfigDict, dict)) and isinstance(val, dict):
                return val  # type: ignore
            if issubclass(self.type, list) and isinstance(val, list):
                # Ensure elements are scalar strings/ints/floats/bools depending on subtype.
                # Assuming YAML structure matches.
                return val  # type: ignore

            return self._convert_value(str(val))

        # 2. Check environment variable
        raw_value = os.getenv(self.name, self.default)
        if isinstance(raw_value, list):
            raw_value = raw_value[0] if raw_value else ""
        return self._convert_value(raw_value)

    def _resolve_wildcard(self) -> List[T]:
        # Gather values from both Persistent Config 'providers' and Env Vars
        values: List[T] = []

        # 1. Extract from PERSISTENT_CONFIG['providers']
        providers = PERSISTENT_CONFIG.get("providers", [])
        if isinstance(providers, list):
            target_type = "openai"
            target_field = "key"

            if "ANTHROPIC" in self.name:
                target_type = "anthropic"

            if "URL" in self.name:
                target_field = "url"

            for p in providers:
                if not isinstance(p, dict):
                    continue
                # Default type is openai
                ptype = p.get("type", "openai").lower()

                if ptype == target_type:
                    val = p.get(target_field)
                    if val:
                        values.append(self._convert_value(str(val)))

        # 2. Extract from Environment Variables
        prefix = self.name.replace("*", "")
        matching_vars = [
            (key, value) for key, value in os.environ.items() if key.startswith(prefix)
        ]

        env_values = [self._convert_value(value) for _, value in sorted(matching_vars)]

        # Combine: Config file providers first, then Env vars.
        all_values = values + env_values

        if not all_values:
            if isinstance(self.default, str) and self.default:
                return [self._convert_value(self.default)]
            elif isinstance(self.default, list):
                return self.default  # type: ignore
            else:
                return []

        return all_values

    def _convert_value(self, value: str) -> T:
        if issubclass(self.type, (StrList, IntList, FloatList, BoolList, ConfigDict)):
            return self.type.from_string(value)  # type: ignore
        elif self.type is str:
            return value  # type: ignore
        elif self.type is int:
            return int(value)  # type: ignore
        elif self.type is float:
            return float(value)  # type: ignore
        elif self.type is bool:
            return value.lower() in ("true", "1", "yes", "on")  # type: ignore
        else:
            return self.type(value)  # type: ignore


# ----------------- OpenGuard Configuration -----------------

OPENGUARD_CONFIG = Config[str](
    name="OPENGUARD_CONFIG",
    type=str,
    default="./guards.yaml",
    description="Path to guards configuration file.",
)

OPENGUARD_OPENAI_URLS = Config[str](
    name="OPENGUARD_OPENAI_URL_*",
    type=str,
    default="http://localhost:11434/v1",
    description=(
        "Downstream OpenAI-compatible API URLs. "
        "Use wildcard naming like OPENGUARD_OPENAI_URL_OLLAMA."
    ),
)

OPENGUARD_OPENAI_KEYS = Config[str](
    name="OPENGUARD_OPENAI_KEY_*",
    type=str,
    default="",
    description=(
        "API keys for downstream APIs. Use wildcard naming like OPENGUARD_OPENAI_KEY_OLLAMA."
    ),
)

OPENGUARD_ANTHROPIC_URLS = Config[str](
    name="OPENGUARD_ANTHROPIC_URL_*",
    type=str,
    default="",
    description=(
        "Downstream Anthropic Chat API URLs. Use wildcard naming like OPENGUARD_ANTHROPIC_URL_1."
    ),
)

OPENGUARD_ANTHROPIC_KEYS = Config[str](
    name="OPENGUARD_ANTHROPIC_KEY_*",
    type=str,
    default="",
    description=(
        "API keys for downstream Anthropic APIs. "
        "Use wildcard naming like OPENGUARD_ANTHROPIC_KEY_1."
    ),
)

OPENGUARD_API_KEY = Config[str](
    name="OPENGUARD_API_KEY",
    type=str,
    default="",
    description="API key required to access this OpenGuard proxy. Leave empty to disable auth.",
)

OPENGUARD_API_KEYS = Config[StrList](
    name="OPENGUARD_API_KEYS",
    type=StrList,
    default="",
    description="Additional API keys (semicolon-separated) for proxy access.",
)

OPENGUARD_PUBLIC_URL = Config[str](
    name="OPENGUARD_PUBLIC_URL",
    type=str,
    default="http://localhost:23294",
    description="Public URL where this OpenGuard instance is accessible.",
)

OPENGUARD_PORT = Config[int](
    name="OPENGUARD_PORT",
    type=int,
    default="23294",
    description="Port to run the OpenGuard server on.",
)

OPENGUARD_HOST = Config[str](
    name="OPENGUARD_HOST",
    type=str,
    default="0.0.0.0",
    description="Host to bind the OpenGuard server to.",
)

OPENGUARD_LOG_LEVEL = Config[str](
    name="OPENGUARD_LOG_LEVEL",
    type=str,
    default="INFO",
    description="Log level (DEBUG, INFO, WARNING, ERROR, CRITICAL).",
)

OPENGUARD_CORS_ORIGINS = Config[StrList](
    name="OPENGUARD_CORS_ORIGINS",
    type=StrList,
    default="*",
    description="Semicolon-separated list of allowed CORS origins.",
)

MODEL_FILTER = Config[ConfigDict](
    name="OPENGUARD_MODEL_FILTER",
    type=ConfigDict,
    default="",
    description="Filter to apply to downstream models list (Hasura-style filter).",
)

INTERMEDIATE_OUTPUT = Config[bool](
    name="BOOST_INTERMEDIATE_OUTPUT",
    type=bool,
    default="false",
    description="Stream intermediate reasoning steps to the client.",
)

EXTRA_LLM_PARAMS = Config[ConfigDict](
    name="BOOST_EXTRA_LLM_PARAMS",
    type=ConfigDict,
    default="",
    description="Extra parameters to inject into every LLM request.",
)

BOOST_PUBLIC_URL = Config[str](
    name="BOOST_PUBLIC_URL",
    type=str,
    default="http://localhost:23294",
    description="Public URL for Boost artifacts.",
)

STATUS_STYLE = Config[str](
    name="STATUS_STYLE",
    type=str,
    default="md:codeblock",
    description="Style for status messages (md:codeblock, md:h1, md:h2, md:h3, plain, none)",
)

if __name__ == "__main__":
    # Render documentation
    configs = [item for item in globals().values() if isinstance(item, Config)]

    # Reload config to ensure clean state
    # PERSISTENT_CONFIG = {}
    _load_persistent_config()

    docs = """
# OpenGuard Configuration

OpenGuard is configured using environment variables or a configuration file at
`~/.config/openguard/config.yaml`.

## Usage
Create `~/.config/openguard/config.yaml` with the following structure:
```yaml
port: 23294
providers:
  - type: openai
    key: sk-proj-...
    url: ...
  - type: anthropic
    key: sk-ant-...
    url: ...

# Other options (match variable names without OPENGUARD_ prefix, lowercase)
model_filter:
  ...
```

Following options are available:
  """

    for config in configs:
        docs += f"\n\n## {config.name}\n"
        docs += f"> **Type**: `{config.type.__name__}`<br/>\n"
        docs += f"> **Default**: `{config.default}`<br/>\n"

        if config.description:
            docs += f"\n{config.description}\n"

    print(docs)



from copy import deepcopy
from typing import Any, Dict, Optional

from . import format, llm, log, selection
from .chat_node import ChatNode

logger = log.setup_logger(__name__)


class Chat:
    tail: ChatNode
    llm: Optional["llm.LLM"]

    @staticmethod
    def from_conversation(messages):
        tail = ChatNode.from_conversation(messages)
        return Chat(tail=tail)

    @staticmethod
    def from_payload(payload: dict, provider: str = "openai"):
        if provider == "anthropic":
            return Chat.from_anthropic_payload(payload)

        messages = payload.get("messages", []) if isinstance(payload, dict) else payload
        return Chat.from_conversation(messages)

    @staticmethod
    def from_anthropic_payload(payload: dict):
        if not isinstance(payload, dict):
            raise ValueError("Anthropic payload must be a dictionary")

        source_payload = deepcopy(payload)
        node_specs = []

        system_content = payload.get("system")
        if isinstance(system_content, str):
            node_specs.append(
                {
                    "role": "system",
                    "content": system_content,
                    "meta": {
                        "anthropic_ref": {
                            "section": "system",
                            "kind": "string",
                        }
                    },
                }
            )
        elif isinstance(system_content, list):
            for block_index, block in enumerate(system_content):
                node_specs.append(
                    Chat._anthropic_block_to_node_spec(block, "system", None, block_index)
                )

        for message_index, message in enumerate(payload.get("messages", [])):
            role = message.get("role", "")
            content = message.get("content", "")

            if isinstance(content, list):
                for block_index, block in enumerate(content):
                    node_specs.append(
                        Chat._anthropic_block_to_node_spec(block, role, message_index, block_index)
                    )
            else:
                node_specs.append(
                    {
                        "role": role,
                        "content": content,
                        "meta": {
                            "anthropic_ref": {
                                "section": "messages",
                                "message_index": message_index,
                                "kind": "message_string",
                            }
                        },
                    }
                )

        tail = ChatNode.from_conversation(
            [
                {
                    "role": spec["role"],
                    "content": spec.get("content", ""),
                    "meta": spec.get("meta", {}),
                }
                for spec in node_specs
            ]
        )

        if tail is not None:
            for node, spec in zip(tail.parents(), node_specs):
                node.meta = spec.get("meta", {})

        return Chat(
            tail=tail,
            provider="anthropic",
            source_payload=source_payload,
        )

    @staticmethod
    def from_tail(chat: "Chat"):
        new_tail = ChatNode(role=chat.tail.role, content=chat.tail.content)
        return Chat(tail=new_tail)

    def __init__(self, **kwargs):
        self.tail = kwargs.get("tail")
        self.llm = kwargs.get("llm")
        self.provider = kwargs.get("provider", "openai")
        self.source_payload = kwargs.get("source_payload", None)
        self.chat_node_type = ChatNode

        self.Chat = Chat
        self.ChatNode = self.chat_node_type

    def clone(self):
        if self.provider == "anthropic":
            return Chat.from_payload(self.serialize("anthropic"), provider="anthropic")
        return Chat.from_conversation(self.history())

    def has_substring(self, substring):
        return any(substring in msg.content for msg in self.plain())

    def match(self, **kwargs):
        return selection.match(self, **kwargs)

    def match_one(self, **kwargs):
        candidates = self.match(**kwargs)
        if not candidates:
            return None
        return candidates[0]

    def add_message(self, role, content):
        logger.debug(f"Chat message: {role}: {content[:50]}")

        child = self.__create_node(role=role, content=content)
        if self.tail:
            self.tail.add_child(child)
        self.tail = child

        return self.tail

    def tool_call(self, tool_call):
        self.add_message("assistant", "")
        self.tail.tool_calls = [tool_call]
        return self.tail

    def tool(self, id, content):
        self.add_message("tool", content)
        self.tail.tool_call_id = id
        return self.tail

    def user(self, content):
        return self.add_message("user", content)

    def assistant(self, content):
        return self.add_message("assistant", content)

    def system(self, content):
        self.tail.ancestor().add_parent(self.__create_node(role="system", content=content))
        return self.tail

    def insert(self, after: ChatNode, role, content):
        new_node = self.__create_node(role=role, content=content)
        after.insert_child(new_node)

        if self.tail == after:
            self.tail = new_node

        return self.tail

    def plain(self):
        if not self.tail:
            return []
        return self.tail.parents()

    def history(self):
        if not self.tail:
            return []
        return self.tail.history()

    def serialize(self, provider: Optional[str] = None):
        active_provider = provider or self.provider or "openai"

        if active_provider == "anthropic":
            return self._serialize_anthropic()

        return self.history()

    def serialized(self, provider: Optional[str] = None):
        return self.serialize(provider=provider)

    def root(self):
        if not self.tail:
            return None
        return self.tail.ancestor()

    def text(self):
        # __str__ already does exactly this
        return f"{self}"

    def __create_node(self, **kwargs):
        NodeType = self.chat_node_type
        return NodeType(**kwargs)

    async def advance(self):
        """
        Advance the chat completion

        Will not be streamed back to the client
        """

        if not self.llm:
            raise ValueError("Chat: unable to advance without an LLM")

        response = await self.llm.chat_completion(chat=self)
        params = await self.llm.resolve_request_params()
        self.assistant(self.llm.get_response_content(params, response))
        return response

    async def emit_advance(self, **kwargs):
        """
        Emit the next step in the chat completion

        Will be streamed back to the client
        """

        if not self.llm:
            raise ValueError("Chat: unable to advance without an LLM")

        response = await self.llm.stream_chat_completion(chat=self, **kwargs)
        self.assistant(response)
        return response

    async def emit_status(self, status):
        """
        Emit a status message

        Will be streamed back to the client
        """

        if not self.llm:
            raise ValueError("Chat: unable to emit status without an LLM")

        await self.llm.emit_status(status)

    async def sanitise_artifacts(self):
        tail = self.tail

        while tail:
            tail_content = tail.content
            tail.content = format.remove_html_code_blocks(tail_content)
            tail = tail.parent

    def __str__(self):
        return "\n".join([str(msg) for msg in self.plain()])

    @staticmethod
    def _anthropic_block_to_node_spec(
        block: Any,
        message_role: str,
        message_index: Optional[int],
        block_index: int,
    ) -> Dict[str, Any]:
        block_type = block.get("type") if isinstance(block, dict) else None
        ref = {
            "section": "messages" if message_index is not None else "system",
            "message_index": message_index,
            "block_index": block_index,
            "kind": "block",
        }

        if block_type == "tool_use":
            return {
                "role": "tool_use",
                "content": deepcopy(block.get("input", {})),
                "meta": {
                    "anthropic_ref": ref,
                    "anthropic_block_type": "tool_use",
                    "anthropic_parent_role": message_role,
                    "anthropic_tool_use_id": block.get("id"),
                    "anthropic_tool_name": block.get("name"),
                },
            }

        if block_type == "tool_result":
            return {
                "role": "tool_result",
                "content": deepcopy(block.get("content", "")),
                "meta": {
                    "anthropic_ref": ref,
                    "anthropic_block_type": "tool_result",
                    "anthropic_parent_role": message_role,
                    "anthropic_tool_use_id": block.get("tool_use_id"),
                    "anthropic_is_error": block.get("is_error"),
                },
            }

        if block_type == "text":
            node_content = block.get("text", "")
        else:
            node_content = deepcopy(block)

        return {
            "role": message_role,
            "content": node_content,
            "meta": {
                "anthropic_ref": ref,
                "anthropic_block_type": block_type,
                "anthropic_parent_role": message_role,
            },
        }

    @staticmethod
    def _node_to_anthropic_block(node: ChatNode, existing: Optional[dict] = None) -> dict:
        block = deepcopy(existing) if isinstance(existing, dict) else {}
        block_type = node.meta.get("anthropic_block_type")

        if node.role == "tool_use" or block_type == "tool_use":
            block["type"] = "tool_use"
            if node.meta.get("anthropic_tool_use_id") is not None:
                block["id"] = node.meta.get("anthropic_tool_use_id")
            if node.meta.get("anthropic_tool_name") is not None:
                block["name"] = node.meta.get("anthropic_tool_name")
            block["input"] = deepcopy(node.content)
            return block

        if node.role == "tool_result" or block_type == "tool_result":
            block["type"] = "tool_result"
            if node.meta.get("anthropic_tool_use_id") is not None:
                block["tool_use_id"] = node.meta.get("anthropic_tool_use_id")
            if node.meta.get("anthropic_is_error") is not None:
                block["is_error"] = node.meta.get("anthropic_is_error")
            block["content"] = deepcopy(node.content)
            return block

        if block_type == "text" or isinstance(node.content, str):
            block["type"] = "text"
            block["text"] = node.content if isinstance(node.content, str) else str(node.content)
            return block

        if isinstance(node.content, dict):
            return deepcopy(node.content)

        block["type"] = block_type or "text"
        if block["type"] == "text":
            block["text"] = node.content if isinstance(node.content, str) else str(node.content)
        else:
            block["content"] = deepcopy(node.content)
        return block

    def _serialize_anthropic(self) -> dict:
        payload = deepcopy(self.source_payload) if isinstance(self.source_payload, dict) else {}
        payload.setdefault("messages", [])

        unmapped_nodes = []

        for node in self.plain():
            ref = node.meta.get("anthropic_ref", {})
            section = ref.get("section")
            kind = ref.get("kind")

            if not section:
                unmapped_nodes.append(node)
                continue

            if section == "system":
                if kind == "string":
                    payload["system"] = node.content
                    continue

                block_index = ref.get("block_index")
                system_value = payload.get("system", [])
                if not isinstance(system_value, list):
                    system_value = []
                while len(system_value) <= block_index:
                    system_value.append({"type": "text", "text": ""})
                existing_block = system_value[block_index]
                system_value[block_index] = self._node_to_anthropic_block(node, existing_block)
                payload["system"] = system_value
                continue

            if section == "messages":
                message_index = ref.get("message_index")
                while len(payload["messages"]) <= message_index:
                    payload["messages"].append({"role": "user", "content": []})

                message = payload["messages"][message_index]
                if kind == "message_string":
                    message["content"] = node.content
                    if node.role:
                        message["role"] = node.role
                    continue

                block_index = ref.get("block_index")
                message_content = message.get("content", [])
                if not isinstance(message_content, list):
                    message_content = []
                while len(message_content) <= block_index:
                    message_content.append({"type": "text", "text": ""})

                existing_block = message_content[block_index]
                message_content[block_index] = self._node_to_anthropic_block(node, existing_block)
                if node.meta.get("anthropic_parent_role"):
                    message["role"] = node.meta.get("anthropic_parent_role")
                elif node.role in ("user", "assistant"):
                    message["role"] = node.role
                message["content"] = message_content
                continue

            unmapped_nodes.append(node)

        for node in unmapped_nodes:
            if node.role == "system":
                system_value = payload.get("system")
                if system_value is None:
                    payload["system"] = node.content
                elif isinstance(system_value, str):
                    payload["system"] = [
                        {"type": "text", "text": system_value},
                        {
                            "type": "text",
                            "text": (
                                node.content if isinstance(node.content, str) else str(node.content)
                            ),
                        },
                    ]
                elif isinstance(system_value, list):
                    system_value.append(self._node_to_anthropic_block(node))
                continue

            if node.role in ("tool_use", "tool_result"):
                role = node.meta.get("anthropic_parent_role")
                if role is None:
                    role = "assistant" if node.role == "tool_use" else "user"
                payload["messages"].append(
                    {
                        "role": role,
                        "content": [self._node_to_anthropic_block(node)],
                    }
                )
                continue

            if isinstance(node.content, list):
                content_value = deepcopy(node.content)
            elif isinstance(node.content, str):
                content_value = [{"type": "text", "text": node.content}]
            else:
                content_value = [self._node_to_anthropic_block(node)]

            payload["messages"].append(
                {
                    "role": node.role if node.role in ("user", "assistant") else "user",
                    "content": content_value,
                }
            )

        return payload
